from graphql import Undefined
from graphql.type import (
    GraphQLArgument,
    GraphQLEnumType,
    GraphQLEnumValue,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputObjectType,
    GraphQLInterfaceType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLString,
)

from ..dynamic import Dynamic
from ..enum import Enum
from ..field import Field
from ..inputfield import InputField
from ..inputobjecttype import InputObjectType
from ..interface import Interface
from ..objecttype import ObjectType
from ..scalars import Int, String
from ..schema import Schema
from ..structures import List, NonNull


def create_type_map(types, auto_camelcase=True):
    query = type("Query", (ObjectType,), {})
    schema = Schema(query, types=types, auto_camelcase=auto_camelcase)
    return schema.graphql_schema.type_map


def test_enum():
    class MyEnum(Enum):
        """Description"""

        foo = 1
        bar = 2

        @property
        def description(self):
            return f"Description {self.name}={self.value}"

        @property
        def deprecation_reason(self):
            if self == MyEnum.foo:
                return "Is deprecated"

    type_map = create_type_map([MyEnum])
    assert "MyEnum" in type_map
    graphql_enum = type_map["MyEnum"]
    assert isinstance(graphql_enum, GraphQLEnumType)
    assert graphql_enum.name == "MyEnum"
    assert graphql_enum.description == "Description"
    assert graphql_enum.values == {
        "foo": GraphQLEnumValue(
            value=1, description="Description foo=1", deprecation_reason="Is deprecated"
        ),
        "bar": GraphQLEnumValue(value=2, description="Description bar=2"),
    }


def test_objecttype():
    class MyObjectType(ObjectType):
        """Description"""

        foo = String(
            bar=String(description="Argument description", default_value="x"),
            description="Field description",
        )
        bar = String(name="gizmo")

        def resolve_foo(self, bar):
            return bar

    type_map = create_type_map([MyObjectType])
    assert "MyObjectType" in type_map
    graphql_type = type_map["MyObjectType"]
    assert isinstance(graphql_type, GraphQLObjectType)
    assert graphql_type.name == "MyObjectType"
    assert graphql_type.description == "Description"

    fields = graphql_type.fields
    assert list(fields) == ["foo", "gizmo"]
    foo_field = fields["foo"]
    assert isinstance(foo_field, GraphQLField)
    assert foo_field.description == "Field description"

    assert foo_field.args == {
        "bar": GraphQLArgument(
            GraphQLString,
            description="Argument description",
            default_value="x",
            out_name="bar",
        )
    }


def test_required_argument_with_default_value():
    class MyObjectType(ObjectType):
        foo = String(bar=String(required=True, default_value="x"))

    type_map = create_type_map([MyObjectType])

    graphql_type = type_map["MyObjectType"]
    foo_field = graphql_type.fields["foo"]

    bar_argument = foo_field.args["bar"]
    assert bar_argument.default_value == "x"
    assert isinstance(bar_argument.type, GraphQLNonNull)
    assert bar_argument.type.of_type == GraphQLString


def test_dynamic_objecttype():
    class MyObjectType(ObjectType):
        """Description"""

        bar = Dynamic(lambda: Field(String))
        own = Field(lambda: MyObjectType)

    type_map = create_type_map([MyObjectType])
    assert "MyObjectType" in type_map
    assert list(MyObjectType._meta.fields) == ["bar", "own"]
    graphql_type = type_map["MyObjectType"]

    fields = graphql_type.fields
    assert list(fields) == ["bar", "own"]
    assert fields["bar"].type == GraphQLString
    assert fields["own"].type == graphql_type


def test_interface():
    class MyInterface(Interface):
        """Description"""

        foo = String(
            bar=String(description="Argument description", default_value="x"),
            description="Field description",
        )
        bar = String(name="gizmo", first_arg=String(), other_arg=String(name="oth_arg"))
        own = Field(lambda: MyInterface)

        def resolve_foo(self, args, info):
            return args.get("bar")

    type_map = create_type_map([MyInterface])
    assert "MyInterface" in type_map
    graphql_type = type_map["MyInterface"]
    assert isinstance(graphql_type, GraphQLInterfaceType)
    assert graphql_type.name == "MyInterface"
    assert graphql_type.description == "Description"

    fields = graphql_type.fields
    assert list(fields) == ["foo", "gizmo", "own"]
    assert fields["own"].type == graphql_type
    assert list(fields["gizmo"].args) == ["firstArg", "oth_arg"]
    foo_field = fields["foo"]
    assert isinstance(foo_field, GraphQLField)
    assert foo_field.description == "Field description"
    assert not foo_field.resolve  # Resolver not attached in interfaces
    assert foo_field.args == {
        "bar": GraphQLArgument(
            GraphQLString,
            description="Argument description",
            default_value="x",
            out_name="bar",
        )
    }


def test_inputobject():
    class OtherObjectType(InputObjectType):
        thingy = NonNull(Int)

    class MyInnerObjectType(InputObjectType):
        some_field = String()
        some_other_field = List(OtherObjectType)

    class MyInputObjectType(InputObjectType):
        """Description"""

        foo_bar = String(description="Field description")
        bar = String(name="gizmo")
        baz = NonNull(MyInnerObjectType)
        own = InputField(lambda: MyInputObjectType)

        def resolve_foo_bar(self, args, info):
            return args.get("bar")

    type_map = create_type_map([MyInputObjectType])
    assert "MyInputObjectType" in type_map
    graphql_type = type_map["MyInputObjectType"]
    assert isinstance(graphql_type, GraphQLInputObjectType)
    assert graphql_type.name == "MyInputObjectType"
    assert graphql_type.description == "Description"

    other_graphql_type = type_map["OtherObjectType"]
    inner_graphql_type = type_map["MyInnerObjectType"]
    container = graphql_type.out_type(
        {
            "bar": "oh!",
            "baz": inner_graphql_type.out_type(
                {
                    "some_other_field": [
                        other_graphql_type.out_type({"thingy": 1}),
                        other_graphql_type.out_type({"thingy": 2}),
                    ]
                }
            ),
        }
    )
    assert isinstance(container, MyInputObjectType)
    assert "bar" in container
    assert container.bar == "oh!"
    assert "foo_bar" not in container
    assert container.foo_bar is None
    assert container.baz.some_field is None
    assert container.baz.some_other_field[0].thingy == 1
    assert container.baz.some_other_field[1].thingy == 2

    fields = graphql_type.fields
    assert list(fields) == ["fooBar", "gizmo", "baz", "own"]
    own_field = fields["own"]
    assert own_field.type == graphql_type
    foo_field = fields["fooBar"]
    assert isinstance(foo_field, GraphQLInputField)
    assert foo_field.description == "Field description"


def test_inputobject_undefined(set_default_input_object_type_to_undefined):
    class OtherObjectType(InputObjectType):
        optional_field = String()

    type_map = create_type_map([OtherObjectType])
    assert "OtherObjectType" in type_map
    graphql_type = type_map["OtherObjectType"]

    container = graphql_type.out_type({})
    assert container.optional_field is Undefined


def test_objecttype_camelcase():
    class MyObjectType(ObjectType):
        """Description"""

        foo_bar = String(bar_foo=String())

    type_map = create_type_map([MyObjectType])
    assert "MyObjectType" in type_map
    graphql_type = type_map["MyObjectType"]
    assert isinstance(graphql_type, GraphQLObjectType)
    assert graphql_type.name == "MyObjectType"
    assert graphql_type.description == "Description"

    fields = graphql_type.fields
    assert list(fields) == ["fooBar"]
    foo_field = fields["fooBar"]
    assert isinstance(foo_field, GraphQLField)
    assert foo_field.args == {
        "barFoo": GraphQLArgument(
            GraphQLString, default_value=Undefined, out_name="bar_foo"
        )
    }


def test_objecttype_camelcase_disabled():
    class MyObjectType(ObjectType):
        """Description"""

        foo_bar = String(bar_foo=String())

    type_map = create_type_map([MyObjectType], auto_camelcase=False)
    assert "MyObjectType" in type_map
    graphql_type = type_map["MyObjectType"]
    assert isinstance(graphql_type, GraphQLObjectType)
    assert graphql_type.name == "MyObjectType"
    assert graphql_type.description == "Description"

    fields = graphql_type.fields
    assert list(fields) == ["foo_bar"]
    foo_field = fields["foo_bar"]
    assert isinstance(foo_field, GraphQLField)
    assert foo_field.args == {
        "bar_foo": GraphQLArgument(
            GraphQLString, default_value=Undefined, out_name="bar_foo"
        )
    }


def test_objecttype_with_possible_types():
    class MyObjectType(ObjectType):
        """Description"""

        class Meta:
            possible_types = (dict,)

        foo_bar = String()

    type_map = create_type_map([MyObjectType])
    graphql_type = type_map["MyObjectType"]
    assert graphql_type.is_type_of
    assert graphql_type.is_type_of({}, None) is True
    assert graphql_type.is_type_of(MyObjectType(), None) is False


def test_interface_with_interfaces():
    class FooInterface(Interface):
        foo = String()

    class BarInterface(Interface):
        class Meta:
            interfaces = [FooInterface]

        foo = String()
        bar = String()

    type_map = create_type_map([FooInterface, BarInterface])
    assert "FooInterface" in type_map
    foo_graphql_type = type_map["FooInterface"]
    assert isinstance(foo_graphql_type, GraphQLInterfaceType)
    assert foo_graphql_type.name == "FooInterface"

    assert "BarInterface" in type_map
    bar_graphql_type = type_map["BarInterface"]
    assert isinstance(bar_graphql_type, GraphQLInterfaceType)
    assert bar_graphql_type.name == "BarInterface"

    fields = bar_graphql_type.fields
    assert list(fields) == ["foo", "bar"]
    assert isinstance(fields["foo"], GraphQLField)
    assert isinstance(fields["bar"], GraphQLField)

    assert list(bar_graphql_type.interfaces) == list([foo_graphql_type])



import re
from textwrap import dedent

from graphql_relay import to_global_id

from ...types import ObjectType, Schema, String
from ..node import Node, is_node


class SharedNodeFields:
    shared = String()
    something_else = String()

    def resolve_something_else(*_):
        return "----"


class MyNode(ObjectType):
    class Meta:
        interfaces = (Node,)

    name = String()

    @staticmethod
    def get_node(info, id):
        return MyNode(name=str(id))


class MyOtherNode(SharedNodeFields, ObjectType):
    extra_field = String()

    class Meta:
        interfaces = (Node,)

    def resolve_extra_field(self, *_):
        return "extra field info."

    @staticmethod
    def get_node(info, id):
        return MyOtherNode(shared=str(id))


class RootQuery(ObjectType):
    first = String()
    node = Node.Field()
    only_node = Node.Field(MyNode)
    only_node_lazy = Node.Field(lambda: MyNode)


schema = Schema(query=RootQuery, types=[MyNode, MyOtherNode])


def test_node_good():
    assert "id" in MyNode._meta.fields
    assert is_node(MyNode)
    assert not is_node(object)
    assert not is_node("node")


def test_node_query():
    executed = schema.execute(
        '{ node(id:"%s") { ... on MyNode { name } } }' % Node.to_global_id("MyNode", 1)
    )
    assert not executed.errors
    assert executed.data == {"node": {"name": "1"}}


def test_subclassed_node_query():
    executed = schema.execute(
        '{ node(id:"%s") { ... on MyOtherNode { shared, extraField, somethingElse } } }'
        % to_global_id("MyOtherNode", 1)
    )
    assert not executed.errors
    assert executed.data == {
        "node": {
            "shared": "1",
            "extraField": "extra field info.",
            "somethingElse": "----",
        }
    }


def test_node_requesting_non_node():
    executed = schema.execute(
        '{ node(id:"%s") { __typename } } ' % Node.to_global_id("RootQuery", 1)
    )
    assert executed.errors
    assert re.match(
        r"ObjectType .* does not implement the .* interface.",
        executed.errors[0].message,
    )
    assert executed.data == {"node": None}


def test_node_requesting_unknown_type():
    executed = schema.execute(
        '{ node(id:"%s") { __typename } } ' % Node.to_global_id("UnknownType", 1)
    )
    assert executed.errors
    assert re.match(r"Relay Node .* not found in schema", executed.errors[0].message)
    assert executed.data == {"node": None}


def test_node_query_incorrect_id():
    executed = schema.execute(
        '{ node(id:"%s") { ... on MyNode { name } } }' % "something:2"
    )
    assert executed.errors
    assert re.match(r"Unable to parse global ID .*", executed.errors[0].message)
    assert executed.data == {"node": None}


def test_node_field():
    node_field = Node.Field()
    assert node_field.type == Node
    assert node_field.node_type == Node


def test_node_field_custom():
    node_field = Node.Field(MyNode)
    assert node_field.type == MyNode
    assert node_field.node_type == Node


def test_node_field_args():
    field_args = {
        "name": "my_custom_name",
        "description": "my_custom_description",
        "deprecation_reason": "my_custom_deprecation_reason",
    }
    node_field = Node.Field(**field_args)
    for field_arg, value in field_args.items():
        assert getattr(node_field, field_arg) == value


def test_node_field_only_type():
    executed = schema.execute(
        '{ onlyNode(id:"%s") { __typename, name } } ' % Node.to_global_id("MyNode", 1)
    )
    assert not executed.errors
    assert executed.data == {"onlyNode": {"__typename": "MyNode", "name": "1"}}


def test_node_field_only_type_wrong():
    executed = schema.execute(
        '{ onlyNode(id:"%s") { __typename, name } } '
        % Node.to_global_id("MyOtherNode", 1)
    )
    assert len(executed.errors) == 1
    assert str(executed.errors[0]).startswith("Must receive a MyNode id.")
    assert executed.data == {"onlyNode": None}


def test_node_field_only_lazy_type():
    executed = schema.execute(
        '{ onlyNodeLazy(id:"%s") { __typename, name } } '
        % Node.to_global_id("MyNode", 1)
    )
    assert not executed.errors
    assert executed.data == {"onlyNodeLazy": {"__typename": "MyNode", "name": "1"}}


def test_node_field_only_lazy_type_wrong():
    executed = schema.execute(
        '{ onlyNodeLazy(id:"%s") { __typename, name } } '
        % Node.to_global_id("MyOtherNode", 1)
    )
    assert len(executed.errors) == 1
    assert str(executed.errors[0]).startswith("Must receive a MyNode id.")
    assert executed.data == {"onlyNodeLazy": None}


def test_str_schema():
    assert (
        str(schema).strip()
        == dedent(
            '''
        schema {
          query: RootQuery
        }

        type MyNode implements Node {
          """The ID of the object"""
          id: ID!
          name: String
        }

        """An object with an ID"""
        interface Node {
          """The ID of the object"""
          id: ID!
        }

        type MyOtherNode implements Node {
          """The ID of the object"""
          id: ID!
          shared: String
          somethingElse: String
          extraField: String
        }

        type RootQuery {
          first: String
          node(
            """The ID of the object"""
            id: ID!
          ): Node
          onlyNode(
            """The ID of the object"""
            id: ID!
          ): MyNode
          onlyNodeLazy(
            """The ID of the object"""
            id: ID!
          ): MyNode
        }
        '''
        ).strip()
    )



from functools import partial

from pytest import raises

from ..argument import Argument
from ..field import Field
from ..scalars import String
from ..structures import NonNull
from .utils import MyLazyType


class MyInstance:
    value = "value"
    value_func = staticmethod(lambda: "value_func")

    def value_method(self):
        return "value_method"


def test_field_basic():
    MyType = object()
    args = {"my arg": Argument(True)}

    def resolver():
        return None

    deprecation_reason = "Deprecated now"
    description = "My Field"
    my_default = "something"
    field = Field(
        MyType,
        name="name",
        args=args,
        resolver=resolver,
        description=description,
        deprecation_reason=deprecation_reason,
        default_value=my_default,
    )
    assert field.name == "name"
    assert field.args == args
    assert field.resolver == resolver
    assert field.deprecation_reason == deprecation_reason
    assert field.description == description
    assert field.default_value == my_default


def test_field_required():
    MyType = object()
    field = Field(MyType, required=True)
    assert isinstance(field.type, NonNull)
    assert field.type.of_type == MyType


def test_field_default_value_not_callable():
    MyType = object()
    try:
        Field(MyType, default_value=lambda: True)
    except AssertionError as e:
        # substring comparison for py 2/3 compatibility
        assert "The default value can not be a function but received" in str(e)


def test_field_source():
    MyType = object()
    field = Field(MyType, source="value")
    assert field.resolver(MyInstance(), None) == MyInstance.value


def test_field_source_dict_or_attr():
    MyType = object()
    field = Field(MyType, source="value")
    assert field.resolver(MyInstance(), None) == MyInstance.value
    assert field.resolver({"value": MyInstance.value}, None) == MyInstance.value


def test_field_with_lazy_type():
    MyType = object()
    field = Field(lambda: MyType)
    assert field.type == MyType


def test_field_with_lazy_partial_type():
    MyType = object()
    field = Field(partial(lambda: MyType))
    assert field.type == MyType


def test_field_with_string_type():
    field = Field("graphene.types.tests.utils.MyLazyType")
    assert field.type == MyLazyType


def test_field_not_source_and_resolver():
    MyType = object()
    with raises(Exception) as exc_info:
        Field(MyType, source="value", resolver=lambda: None)
    assert (
        str(exc_info.value)
        == "A Field cannot have a source and a resolver in at the same time."
    )


def test_field_source_func():
    MyType = object()
    field = Field(MyType, source="value_func")
    assert field.resolver(MyInstance(), None) == MyInstance.value_func()


def test_field_source_method():
    MyType = object()
    field = Field(MyType, source="value_method")
    assert field.resolver(MyInstance(), None) == MyInstance().value_method()


def test_field_source_as_argument():
    MyType = object()
    field = Field(MyType, source=String())
    assert "source" in field.args
    assert field.args["source"].type == String


def test_field_name_as_argument():
    MyType = object()
    field = Field(MyType, name=String())
    assert "name" in field.args
    assert field.args["name"].type == String


def test_field_source_argument_as_kw():
    MyType = object()
    deprecation_reason = "deprecated"
    field = Field(
        MyType,
        b=NonNull(True),
        c=Argument(None, deprecation_reason=deprecation_reason),
        a=NonNull(False),
    )
    assert list(field.args) == ["b", "c", "a"]
    assert isinstance(field.args["b"], Argument)
    assert isinstance(field.args["b"].type, NonNull)
    assert field.args["b"].type.of_type is True
    assert isinstance(field.args["c"], Argument)
    assert field.args["c"].type is None
    assert field.args["c"].deprecation_reason == deprecation_reason
    assert isinstance(field.args["a"], Argument)
    assert isinstance(field.args["a"].type, NonNull)
    assert field.args["a"].type.of_type is False



from ..base import BaseOptions, BaseType


class CustomOptions(BaseOptions):
    pass


class CustomType(BaseType):
    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        _meta = CustomOptions(cls)
        super(CustomType, cls).__init_subclass_with_meta__(_meta=_meta, **options)


def test_basetype():
    class MyBaseType(CustomType):
        pass

    assert isinstance(MyBaseType._meta, CustomOptions)
    assert MyBaseType._meta.name == "MyBaseType"
    assert MyBaseType._meta.description is None


def test_basetype_nones():
    class MyBaseType(CustomType):
        """Documentation"""

        class Meta:
            name = None
            description = None

    assert isinstance(MyBaseType._meta, CustomOptions)
    assert MyBaseType._meta.name == "MyBaseType"
    assert MyBaseType._meta.description == "Documentation"


def test_basetype_custom():
    class MyBaseType(CustomType):
        """Documentation"""

        class Meta:
            name = "Base"
            description = "Desc"

    assert isinstance(MyBaseType._meta, CustomOptions)
    assert MyBaseType._meta.name == "Base"
    assert MyBaseType._meta.description == "Desc"


def test_basetype_create():
    MyBaseType = CustomType.create_type("MyBaseType")

    assert isinstance(MyBaseType._meta, CustomOptions)
    assert MyBaseType._meta.name == "MyBaseType"
    assert MyBaseType._meta.description is None


def test_basetype_create_extra():
    MyBaseType = CustomType.create_type("MyBaseType", name="Base", description="Desc")

    assert isinstance(MyBaseType._meta, CustomOptions)
    assert MyBaseType._meta.name == "Base"
    assert MyBaseType._meta.description == "Desc"



from textwrap import dedent

from ..argument import Argument
from ..enum import Enum, PyEnum
from ..field import Field
from ..inputfield import InputField
from ..inputobjecttype import InputObjectType
from ..mutation import Mutation
from ..scalars import String
from ..schema import ObjectType, Schema


def test_enum_construction():
    class RGB(Enum):
        """Description"""

        RED = 1
        GREEN = 2
        BLUE = 3

        @property
        def description(self):
            return f"Description {self.name}"

    assert RGB._meta.name == "RGB"
    assert RGB._meta.description == "Description"

    values = RGB._meta.enum.__members__.values()
    assert sorted(v.name for v in values) == ["BLUE", "GREEN", "RED"]
    assert sorted(v.description for v in values) == [
        "Description BLUE",
        "Description GREEN",
        "Description RED",
    ]


def test_enum_construction_meta():
    class RGB(Enum):
        class Meta:
            name = "RGBEnum"
            description = "Description"

        RED = 1
        GREEN = 2
        BLUE = 3

    assert RGB._meta.name == "RGBEnum"
    assert RGB._meta.description == "Description"


def test_enum_instance_construction():
    RGB = Enum("RGB", "RED,GREEN,BLUE")

    values = RGB._meta.enum.__members__.values()
    assert sorted(v.name for v in values) == ["BLUE", "GREEN", "RED"]


def test_enum_from_builtin_enum():
    PyRGB = PyEnum("RGB", "RED,GREEN,BLUE")

    RGB = Enum.from_enum(PyRGB)
    assert RGB._meta.enum == PyRGB
    assert RGB.RED
    assert RGB.GREEN
    assert RGB.BLUE


def test_enum_custom_description_in_constructor():
    description = "An enumeration, but with a custom description"
    RGB = Enum(
        "RGB",
        "RED,GREEN,BLUE",
        description=description,
    )
    assert RGB._meta.description == description


def test_enum_from_python3_enum_uses_default_builtin_doc():
    RGB = Enum("RGB", "RED,GREEN,BLUE")
    assert RGB._meta.description == "An enumeration."


def test_enum_from_builtin_enum_accepts_lambda_description():
    def custom_description(value):
        if not value:
            return "StarWars Episodes"

        return "New Hope Episode" if value == Episode.NEWHOPE else "Other"

    def custom_deprecation_reason(value):
        return "meh" if value == Episode.NEWHOPE else None

    PyEpisode = PyEnum("PyEpisode", "NEWHOPE,EMPIRE,JEDI")
    Episode = Enum.from_enum(
        PyEpisode,
        description=custom_description,
        deprecation_reason=custom_deprecation_reason,
    )

    class Query(ObjectType):
        foo = Episode()

    schema = Schema(query=Query).graphql_schema

    episode = schema.get_type("PyEpisode")

    assert episode.description == "StarWars Episodes"
    assert [
        (name, value.description, value.deprecation_reason)
        for name, value in episode.values.items()
    ] == [
        ("NEWHOPE", "New Hope Episode", "meh"),
        ("EMPIRE", "Other", None),
        ("JEDI", "Other", None),
    ]


def test_enum_from_python3_enum_uses_enum_doc():
    from enum import Enum as PyEnum

    class Color(PyEnum):
        """This is the description"""

        RED = 1
        GREEN = 2
        BLUE = 3

    RGB = Enum.from_enum(Color)
    assert RGB._meta.enum == Color
    assert RGB._meta.description == "This is the description"
    assert RGB
    assert RGB.RED
    assert RGB.GREEN
    assert RGB.BLUE


def test_enum_value_from_class():
    class RGB(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    assert RGB.RED.value == 1
    assert RGB.GREEN.value == 2
    assert RGB.BLUE.value == 3


def test_enum_value_as_unmounted_field():
    class RGB(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    unmounted = RGB()
    unmounted_field = unmounted.Field()
    assert isinstance(unmounted_field, Field)
    assert unmounted_field.type == RGB


def test_enum_value_as_unmounted_inputfield():
    class RGB(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    unmounted = RGB()
    unmounted_field = unmounted.InputField()
    assert isinstance(unmounted_field, InputField)
    assert unmounted_field.type == RGB


def test_enum_value_as_unmounted_argument():
    class RGB(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    unmounted = RGB()
    unmounted_field = unmounted.Argument()
    assert isinstance(unmounted_field, Argument)
    assert unmounted_field.type == RGB


def test_enum_can_be_compared():
    class RGB(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    assert RGB.RED == 1
    assert RGB.GREEN == 2
    assert RGB.BLUE == 3


def test_enum_can_be_initialized():
    class RGB(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    assert RGB.get(1) == RGB.RED
    assert RGB.get(2) == RGB.GREEN
    assert RGB.get(3) == RGB.BLUE


def test_enum_can_retrieve_members():
    class RGB(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    assert RGB["RED"] == RGB.RED
    assert RGB["GREEN"] == RGB.GREEN
    assert RGB["BLUE"] == RGB.BLUE


def test_enum_to_enum_comparison_should_differ():
    class RGB1(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    class RGB2(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    assert RGB1.RED != RGB2.RED
    assert RGB1.GREEN != RGB2.GREEN
    assert RGB1.BLUE != RGB2.BLUE


def test_enum_skip_meta_from_members():
    class RGB1(Enum):
        class Meta:
            name = "RGB"

        RED = 1
        GREEN = 2
        BLUE = 3

    assert dict(RGB1._meta.enum.__members__) == {
        "RED": RGB1.RED,
        "GREEN": RGB1.GREEN,
        "BLUE": RGB1.BLUE,
    }


def test_enum_types():
    from enum import Enum as PyEnum

    class Color(PyEnum):
        """Primary colors"""

        RED = 1
        YELLOW = 2
        BLUE = 3

    GColor = Enum.from_enum(Color)

    class Query(ObjectType):
        color = GColor(required=True)

        def resolve_color(_, info):
            return Color.RED

    schema = Schema(query=Query)

    assert (
        str(schema).strip()
        == dedent(
            '''
            type Query {
              color: Color!
            }

            """Primary colors"""
            enum Color {
              RED
              YELLOW
              BLUE
            }
            '''
        ).strip()
    )


def test_enum_resolver():
    from enum import Enum as PyEnum

    class Color(PyEnum):
        RED = 1
        GREEN = 2
        BLUE = 3

    GColor = Enum.from_enum(Color)

    class Query(ObjectType):
        color = GColor(required=True)

        def resolve_color(_, info):
            return Color.RED

    schema = Schema(query=Query)

    results = schema.execute("query { color }")
    assert not results.errors

    assert results.data["color"] == Color.RED.name


def test_enum_resolver_compat():
    from enum import Enum as PyEnum

    class Color(PyEnum):
        RED = 1
        GREEN = 2
        BLUE = 3

    GColor = Enum.from_enum(Color)

    class Query(ObjectType):
        color = GColor(required=True)
        color_by_name = GColor(required=True)

        def resolve_color(_, info):
            return Color.RED.value

        def resolve_color_by_name(_, info):
            return Color.RED.name

    schema = Schema(query=Query)

    results = schema.execute(
        """query {
            color
            colorByName
        }"""
    )
    assert not results.errors

    assert results.data["color"] == Color.RED.name
    assert results.data["colorByName"] == Color.RED.name


def test_enum_with_name():
    from enum import Enum as PyEnum

    class Color(PyEnum):
        RED = 1
        YELLOW = 2
        BLUE = 3

    GColor = Enum.from_enum(Color, description="original colors")
    UniqueGColor = Enum.from_enum(
        Color, name="UniqueColor", description="unique colors"
    )

    class Query(ObjectType):
        color = GColor(required=True)
        unique_color = UniqueGColor(required=True)

    schema = Schema(query=Query)

    assert (
        str(schema).strip()
        == dedent(
            '''
            type Query {
              color: Color!
              uniqueColor: UniqueColor!
            }

            """original colors"""
            enum Color {
              RED
              YELLOW
              BLUE
            }

            """unique colors"""
            enum UniqueColor {
              RED
              YELLOW
              BLUE
            }
            '''
        ).strip()
    )


def test_enum_resolver_invalid():
    from enum import Enum as PyEnum

    class Color(PyEnum):
        RED = 1
        GREEN = 2
        BLUE = 3

    GColor = Enum.from_enum(Color)

    class Query(ObjectType):
        color = GColor(required=True)

        def resolve_color(_, info):
            return "BLACK"

    schema = Schema(query=Query)

    results = schema.execute("query { color }")
    assert results.errors
    assert results.errors[0].message == "Enum 'Color' cannot represent value: 'BLACK'"


def test_field_enum_argument():
    class Color(Enum):
        RED = 1
        GREEN = 2
        BLUE = 3

    class Brick(ObjectType):
        color = Color(required=True)

    color_filter = None

    class Query(ObjectType):
        bricks_by_color = Field(Brick, color=Color(required=True))

        def resolve_bricks_by_color(_, info, color):
            nonlocal color_filter
            color_filter = color
            return Brick(color=color)

    schema = Schema(query=Query)

    results = schema.execute(
        """
        query {
            bricksByColor(color: RED) {
                color
            }
        }
    """
    )
    assert not results.errors
    assert results.data == {"bricksByColor": {"color": "RED"}}
    assert color_filter == Color.RED


def test_mutation_enum_input():
    class RGB(Enum):
        """Available colors"""

        RED = 1
        GREEN = 2
        BLUE = 3

    color_input = None

    class CreatePaint(Mutation):
        class Arguments:
            color = RGB(required=True)

        color = RGB(required=True)

        def mutate(_, info, color):
            nonlocal color_input
            color_input = color
            return CreatePaint(color=color)

    class MyMutation(ObjectType):
        create_paint = CreatePaint.Field()

    class Query(ObjectType):
        a = String()

    schema = Schema(query=Query, mutation=MyMutation)
    result = schema.execute(
        """ mutation MyMutation {
        createPaint(color: RED) {
            color
        }
    }
    """
    )
    assert not result.errors
    assert result.data == {"createPaint": {"color": "RED"}}

    assert color_input == RGB.RED


def test_mutation_enum_input_type():
    class RGB(Enum):
        """Available colors"""

        RED = 1
        GREEN = 2
        BLUE = 3

    class ColorInput(InputObjectType):
        color = RGB(required=True)

    color_input_value = None

    class CreatePaint(Mutation):
        class Arguments:
            color_input = ColorInput(required=True)

        color = RGB(required=True)

        def mutate(_, info, color_input):
            nonlocal color_input_value
            color_input_value = color_input.color
            return CreatePaint(color=color_input.color)

    class MyMutation(ObjectType):
        create_paint = CreatePaint.Field()

    class Query(ObjectType):
        a = String()

    schema = Schema(query=Query, mutation=MyMutation)
    result = schema.execute(
        """
        mutation MyMutation {
            createPaint(colorInput: { color: RED }) {
                color
            }
        }
        """
    )
    assert not result.errors
    assert result.data == {"createPaint": {"color": "RED"}}

    assert color_input_value == RGB.RED


def test_hashable_enum():
    class RGB(Enum):
        """Available colors"""

        RED = 1
        GREEN = 2
        BLUE = 3

    color_map = {RGB.RED: "a", RGB.BLUE: "b", 1: "c"}

    assert color_map[RGB.RED] == "a"
    assert color_map[RGB.BLUE] == "b"
    assert color_map[1] == "c"


def test_hashable_instance_creation_enum():
    Episode = Enum("Episode", [("NEWHOPE", 4), ("EMPIRE", 5), ("JEDI", 6)])

    trilogy_map = {Episode.NEWHOPE: "better", Episode.EMPIRE: "best", 5: "foo"}

    assert trilogy_map[Episode.NEWHOPE] == "better"
    assert trilogy_map[Episode.EMPIRE] == "best"
    assert trilogy_map[5] == "foo"


def test_enum_iteration():
    class TestEnum(Enum):
        FIRST = 1
        SECOND = 2

    result = []
    expected_values = ["FIRST", "SECOND"]
    for c in TestEnum:
        result.append(c.name)
    assert result == expected_values


def test_iterable_instance_creation_enum():
    TestEnum = Enum("TestEnum", [("FIRST", 1), ("SECOND", 2)])

    result = []
    expected_values = ["FIRST", "SECOND"]
    for c in TestEnum:
        result.append(c.name)
    assert result == expected_values


# https://github.com/graphql-python/graphene/issues/1321
def test_enum_description_member_not_interpreted_as_property():
    class RGB(Enum):
        """Description"""

        red = "red"
        green = "green"
        blue = "blue"
        description = "description"
        deprecation_reason = "deprecation_reason"

    class Query(ObjectType):
        color = RGB()

        def resolve_color(_, info):
            return RGB.description

    values = RGB._meta.enum.__members__.values()
    assert sorted(v.name for v in values) == [
        "blue",
        "deprecation_reason",
        "description",
        "green",
        "red",
    ]

    schema = Schema(query=Query)

    results = schema.execute("query { color }")
    assert not results.errors
    assert results.data["color"] == RGB.description.name



from functools import wraps
from typing_extensions import deprecated


@deprecated("This function is deprecated")
def resolve_only_args(func):
    @wraps(func)
    def wrapped_func(root, info, **args):
        return func(root, **args)

    return wrapped_func



from typing import Any

from graphql import Undefined
from graphql.language.ast import (
    BooleanValueNode,
    FloatValueNode,
    IntValueNode,
    StringValueNode,
)

from .base import BaseOptions, BaseType
from .unmountedtype import UnmountedType


class ScalarOptions(BaseOptions):
    pass


class Scalar(UnmountedType, BaseType):
    """
    Scalar Type Definition

    The leaf values of any request and input values to arguments are
    Scalars (or Enums) and are defined with a name and a series of functions
    used to parse input from ast or variables and to ensure validity.
    """

    @classmethod
    def __init_subclass_with_meta__(cls, **options):
        _meta = ScalarOptions(cls)
        super(Scalar, cls).__init_subclass_with_meta__(_meta=_meta, **options)

    serialize = None
    parse_value = None
    parse_literal = None

    @classmethod
    def get_type(cls):
        """
        This function is called when the unmounted type (Scalar instance)
        is mounted (as a Field, InputField or Argument)
        """
        return cls


# As per the GraphQL Spec, Integers are only treated as valid when a valid
# 32-bit signed integer, providing the broadest support across platforms.
#
# n.b. JavaScript's integers are safe between -(2^53 - 1) and 2^53 - 1 because
# they are internally represented as IEEE 754 doubles.
MAX_INT = 2147483647
MIN_INT = -2147483648


class Int(Scalar):
    """
    The `Int` scalar type represents non-fractional signed whole numeric
    values. Int can represent values between -(2^53 - 1) and 2^53 - 1 since
    represented in JSON as double-precision floating point numbers specified
    by [IEEE 754](http://en.wikipedia.org/wiki/IEEE_floating_point).
    """

    @staticmethod
    def coerce_int(value):
        try:
            num = int(value)
        except ValueError:
            try:
                num = int(float(value))
            except ValueError:
                return Undefined
        if MIN_INT <= num <= MAX_INT:
            return num
        return Undefined

    serialize = coerce_int
    parse_value = coerce_int

    @staticmethod
    def parse_literal(ast, _variables=None):
        if isinstance(ast, IntValueNode):
            num = int(ast.value)
            if MIN_INT <= num <= MAX_INT:
                return num
        return Undefined


class BigInt(Scalar):
    """
    The `BigInt` scalar type represents non-fractional whole numeric values.
    `BigInt` is not constrained to 32-bit like the `Int` type and thus is a less
    compatible type.
    """

    @staticmethod
    def coerce_int(value):
        try:
            num = int(value)
        except ValueError:
            try:
                num = int(float(value))
            except ValueError:
                return Undefined
        return num

    serialize = coerce_int
    parse_value = coerce_int

    @staticmethod
    def parse_literal(ast, _variables=None):
        if isinstance(ast, IntValueNode):
            return int(ast.value)
        return Undefined


class Float(Scalar):
    """
    The `Float` scalar type represents signed double-precision fractional
    values as specified by
    [IEEE 754](http://en.wikipedia.org/wiki/IEEE_floating_point).
    """

    @staticmethod
    def coerce_float(value: Any) -> float:
        try:
            return float(value)
        except ValueError:
            return Undefined

    serialize = coerce_float
    parse_value = coerce_float

    @staticmethod
    def parse_literal(ast, _variables=None):
        if isinstance(ast, (FloatValueNode, IntValueNode)):
            return float(ast.value)
        return Undefined


class String(Scalar):
    """
    The `String` scalar type represents textual data, represented as UTF-8
    character sequences. The String type is most often used by GraphQL to
    represent free-form human-readable text.
    """

    @staticmethod
    def coerce_string(value):
        if isinstance(value, bool):
            return "true" if value else "false"
        return str(value)

    serialize = coerce_string
    parse_value = coerce_string

    @staticmethod
    def parse_literal(ast, _variables=None):
        if isinstance(ast, StringValueNode):
            return ast.value
        return Undefined


class Boolean(Scalar):
    """
    The `Boolean` scalar type represents `true` or `false`.
    """

    serialize = bool
    parse_value = bool

    @staticmethod
    def parse_literal(ast, _variables=None):
        if isinstance(ast, BooleanValueNode):
            return ast.value
        return Undefined


class ID(Scalar):
    """
    The `ID` scalar type represents a unique identifier, often used to
    refetch an object or as key for a cache. The ID type appears in a JSON
    response as a String; however, it is not intended to be human-readable.
    When expected as an input type, any string (such as `"4"`) or integer
    (such as `4`) input value will be accepted as an ID.
    """

    serialize = str
    parse_value = str

    @staticmethod
    def parse_literal(ast, _variables=None):
        if isinstance(ast, (StringValueNode, IntValueNode)):
            return ast.value
        return Undefined



from ..orderedtype import OrderedType


def test_orderedtype():
    one = OrderedType()
    two = OrderedType()
    three = OrderedType()

    assert one < two < three


def test_orderedtype_eq():
    one = OrderedType()
    two = OrderedType()

    assert one == one
    assert one != two


def test_orderedtype_hash():
    one = OrderedType()
    two = OrderedType()

    assert hash(one) == hash(one)
    assert hash(one) != hash(two)


def test_orderedtype_resetcounter():
    one = OrderedType()
    two = OrderedType()
    one.reset_counter()

    assert one > two


def test_orderedtype_non_orderabletypes():
    one = OrderedType()

    assert one.__lt__(1) == NotImplemented
    assert one.__gt__(1) == NotImplemented
    assert one != 1



from .. import deprecated
from ..deprecated import warn_deprecation


def test_warn_deprecation(mocker):
    mocker.patch.object(deprecated, "warn")

    warn_deprecation("OH!")
    deprecated.warn.assert_called_with("OH!", stacklevel=2, category=DeprecationWarning)



from pytest import raises

from graphene import ObjectType, String

from ..module_loading import import_string, lazy_import


def test_import_string():
    MyString = import_string("graphene.String")
    assert MyString == String

    MyObjectTypeMeta = import_string("graphene.ObjectType", "__doc__")
    assert MyObjectTypeMeta == ObjectType.__doc__


def test_import_string_module():
    with raises(Exception) as exc_info:
        import_string("graphenea")

    assert str(exc_info.value) == "graphenea doesn't look like a module path"


def test_import_string_class():
    with raises(Exception) as exc_info:
        import_string("graphene.Stringa")

    assert (
        str(exc_info.value)
        == 'Module "graphene" does not define a "Stringa" attribute/class'
    )


def test_import_string_attributes():
    with raises(Exception) as exc_info:
        import_string("graphene.String", "length")

    assert (
        str(exc_info.value)
        == 'Module "graphene" does not define a "length" attribute inside attribute/class '
        '"String"'
    )

    with raises(Exception) as exc_info:
        import_string("graphene.ObjectType", "__class__.length")

    assert (
        str(exc_info.value)
        == 'Module "graphene" does not define a "__class__.length" attribute inside '
        'attribute/class "ObjectType"'
    )

    with raises(Exception) as exc_info:
        import_string("graphene.ObjectType", "__classa__.__base__")

    assert (
        str(exc_info.value)
        == 'Module "graphene" does not define a "__classa__" attribute inside attribute/class '
        '"ObjectType"'
    )


def test_lazy_import():
    f = lazy_import("graphene.String")
    MyString = f()
    assert MyString == String

    f = lazy_import("graphene.ObjectType", "__doc__")
    MyObjectTypeMeta = f()
    assert MyObjectTypeMeta == ObjectType.__doc__



from graphql import Undefined
from ..scalars import Boolean, Float, Int, String


def test_serializes_output_int():
    assert Int.serialize(1) == 1
    assert Int.serialize(0) == 0
    assert Int.serialize(-1) == -1
    assert Int.serialize(0.1) == 0
    assert Int.serialize(1.1) == 1
    assert Int.serialize(-1.1) == -1
    assert Int.serialize(1e5) == 100000
    assert Int.serialize(9876504321) is Undefined
    assert Int.serialize(-9876504321) is Undefined
    assert Int.serialize(1e100) is Undefined
    assert Int.serialize(-1e100) is Undefined
    assert Int.serialize("-1.1") == -1
    assert Int.serialize("one") is Undefined
    assert Int.serialize(False) == 0
    assert Int.serialize(True) == 1


def test_serializes_output_float():
    assert Float.serialize(1) == 1.0
    assert Float.serialize(0) == 0.0
    assert Float.serialize(-1) == -1.0
    assert Float.serialize(0.1) == 0.1
    assert Float.serialize(1.1) == 1.1
    assert Float.serialize(-1.1) == -1.1
    assert Float.serialize("-1.1") == -1.1
    assert Float.serialize("one") is Undefined
    assert Float.serialize(False) == 0
    assert Float.serialize(True) == 1


def test_serializes_output_string():
    assert String.serialize("string") == "string"
    assert String.serialize(1) == "1"
    assert String.serialize(-1.1) == "-1.1"
    assert String.serialize(True) == "true"
    assert String.serialize(False) == "false"
    assert String.serialize("\U0001f601") == "\U0001f601"


def test_serializes_output_boolean():
    assert Boolean.serialize("string") is True
    assert Boolean.serialize("") is False
    assert Boolean.serialize(1) is True
    assert Boolean.serialize(0) is False
    assert Boolean.serialize(True) is True
    assert Boolean.serialize(False) is False



# https://github.com/graphql-python/graphene/issues/1293

from datetime import datetime, timezone

import graphene
from graphql.utilities import print_schema


class Filters(graphene.InputObjectType):
    datetime_after = graphene.DateTime(
        required=False,
        default_value=datetime.fromtimestamp(1434549820.776, timezone.utc),
    )
    datetime_before = graphene.DateTime(
        required=False,
        default_value=datetime.fromtimestamp(1444549820.776, timezone.utc),
    )


class SetDatetime(graphene.Mutation):
    class Arguments:
        filters = Filters(required=True)

    ok = graphene.Boolean()

    def mutate(root, info, filters):
        return SetDatetime(ok=True)


class Query(graphene.ObjectType):
    goodbye = graphene.String()


class Mutations(graphene.ObjectType):
    set_datetime = SetDatetime.Field()


def test_schema_printable_with_default_datetime_value():
    schema = graphene.Schema(query=Query, mutation=Mutations)
    schema_str = print_schema(schema.graphql_schema)
    assert schema_str, "empty schema printed"



from pytest import raises

from ..field import Field
from ..objecttype import ObjectType
from ..union import Union
from ..unmountedtype import UnmountedType


class MyObjectType1(ObjectType):
    pass


class MyObjectType2(ObjectType):
    pass


def test_generate_union():
    class MyUnion(Union):
        """Documentation"""

        class Meta:
            types = (MyObjectType1, MyObjectType2)

    assert MyUnion._meta.name == "MyUnion"
    assert MyUnion._meta.description == "Documentation"
    assert MyUnion._meta.types == (MyObjectType1, MyObjectType2)


def test_generate_union_with_meta():
    class MyUnion(Union):
        class Meta:
            name = "MyOtherUnion"
            description = "Documentation"
            types = (MyObjectType1, MyObjectType2)

    assert MyUnion._meta.name == "MyOtherUnion"
    assert MyUnion._meta.description == "Documentation"


def test_generate_union_with_no_types():
    with raises(Exception) as exc_info:

        class MyUnion(Union):
            pass

    assert str(exc_info.value) == "Must provide types for Union MyUnion."


def test_union_can_be_mounted():
    class MyUnion(Union):
        class Meta:
            types = (MyObjectType1, MyObjectType2)

    my_union_instance = MyUnion()
    assert isinstance(my_union_instance, UnmountedType)
    my_union_field = my_union_instance.mount_as(Field)
    assert isinstance(my_union_field, Field)
    assert my_union_field.type == MyUnion



from ..field import Field
from ..interface import Interface
from ..objecttype import ObjectType
from ..scalars import String
from ..schema import Schema
from ..unmountedtype import UnmountedType


class MyType:
    pass


class MyScalar(UnmountedType):
    def get_type(self):
        return MyType


def test_generate_interface():
    class MyInterface(Interface):
        """Documentation"""

    assert MyInterface._meta.name == "MyInterface"
    assert MyInterface._meta.description == "Documentation"
    assert MyInterface._meta.fields == {}


def test_generate_interface_with_meta():
    class MyFirstInterface(Interface):
        pass

    class MyInterface(Interface):
        class Meta:
            name = "MyOtherInterface"
            description = "Documentation"
            interfaces = [MyFirstInterface]

    assert MyInterface._meta.name == "MyOtherInterface"
    assert MyInterface._meta.description == "Documentation"
    assert MyInterface._meta.interfaces == [MyFirstInterface]


def test_generate_interface_with_fields():
    class MyInterface(Interface):
        field = Field(MyType)

    assert "field" in MyInterface._meta.fields


def test_ordered_fields_in_interface():
    class MyInterface(Interface):
        b = Field(MyType)
        a = Field(MyType)
        field = MyScalar()
        asa = Field(MyType)

    assert list(MyInterface._meta.fields) == ["b", "a", "field", "asa"]


def test_generate_interface_unmountedtype():
    class MyInterface(Interface):
        field = MyScalar()

    assert "field" in MyInterface._meta.fields
    assert isinstance(MyInterface._meta.fields["field"], Field)


def test_generate_interface_inherit_abstracttype():
    class MyAbstractType:
        field1 = MyScalar()

    class MyInterface(Interface, MyAbstractType):
        field2 = MyScalar()

    assert list(MyInterface._meta.fields) == ["field1", "field2"]
    assert [type(x) for x in MyInterface._meta.fields.values()] == [Field, Field]


def test_generate_interface_inherit_interface():
    class MyBaseInterface(Interface):
        field1 = MyScalar()

    class MyInterface(MyBaseInterface):
        field2 = MyScalar()

    assert MyInterface._meta.name == "MyInterface"
    assert list(MyInterface._meta.fields) == ["field1", "field2"]
    assert [type(x) for x in MyInterface._meta.fields.values()] == [Field, Field]


def test_generate_interface_inherit_abstracttype_reversed():
    class MyAbstractType:
        field1 = MyScalar()

    class MyInterface(MyAbstractType, Interface):
        field2 = MyScalar()

    assert list(MyInterface._meta.fields) == ["field1", "field2"]
    assert [type(x) for x in MyInterface._meta.fields.values()] == [Field, Field]


def test_resolve_type_default():
    class MyInterface(Interface):
        field2 = String()

    class MyTestType(ObjectType):
        class Meta:
            interfaces = (MyInterface,)

    class Query(ObjectType):
        test = Field(MyInterface)

        def resolve_test(_, info):
            return MyTestType()

    schema = Schema(query=Query, types=[MyTestType])

    result = schema.execute(
        """
        query {
            test {
                __typename
            }
        }
    """
    )
    assert not result.errors
    assert result.data == {"test": {"__typename": "MyTestType"}}


def test_resolve_type_custom():
    class MyInterface(Interface):
        field2 = String()

        @classmethod
        def resolve_type(cls, instance, info):
            if instance["type"] == 1:
                return MyTestType1
            return MyTestType2

    class MyTestType1(ObjectType):
        class Meta:
            interfaces = (MyInterface,)

    class MyTestType2(ObjectType):
        class Meta:
            interfaces = (MyInterface,)

    class Query(ObjectType):
        test = Field(MyInterface)

        def resolve_test(_, info):
            return {"type": 1}

    schema = Schema(query=Query, types=[MyTestType1, MyTestType2])

    result = schema.execute(
        """
        query {
            test {
                __typename
            }
        }
    """
    )
    assert not result.errors
    assert result.data == {"test": {"__typename": "MyTestType1"}}


def test_resolve_type_custom_interferes():
    class MyInterface(Interface):
        field2 = String()
        type_ = String(name="type")

        def resolve_type_(_, info):
            return "foo"

    class MyTestType1(ObjectType):
        class Meta:
            interfaces = (MyInterface,)

    class MyTestType2(ObjectType):
        class Meta:
            interfaces = (MyInterface,)

    class Query(ObjectType):
        test = Field(MyInterface)

        def resolve_test(_, info):
            return MyTestType1()

    schema = Schema(query=Query, types=[MyTestType1, MyTestType2])

    result = schema.execute(
        """
        query {
            test {
                __typename
                type
            }
        }
    """
    )
    assert not result.errors
    assert result.data == {"test": {"__typename": "MyTestType1", "type": "foo"}}



from pytest import mark

from graphene import ObjectType, Int, String, Schema, Field


class Query(ObjectType):
    hello = String()

    def resolve_hello(root, info):
        return "Hello, world!"


class Subscription(ObjectType):
    count_to_ten = Field(Int)

    async def subscribe_count_to_ten(root, info):
        for count in range(1, 11):
            yield count


schema = Schema(query=Query, subscription=Subscription)


@mark.asyncio
async def test_subscription():
    subscription = "subscription { countToTen }"
    result = await schema.subscribe(subscription)
    count = 0
    async for item in result:
        count = item.data["countToTen"]
    assert count == 10


@mark.asyncio
async def test_subscription_fails_with_invalid_query():
    # It fails if the provided query is invalid
    subscription = "subscription { "
    result = await schema.subscribe(subscription)
    assert not result.data
    assert result.errors
    assert "Syntax Error: Expected Name, found <EOF>" in str(result.errors[0])


@mark.asyncio
async def test_subscription_fails_when_query_is_not_valid():
    # It can't subscribe to two fields at the same time, triggering a
    # validation error.
    subscription = "subscription { countToTen, b: countToTen }"
    result = await schema.subscribe(subscription)
    assert not result.data
    assert result.errors
    assert "Anonymous Subscription must select only one top level field." in str(
        result.errors[0]
    )


@mark.asyncio
async def test_subscription_with_args():
    class Query(ObjectType):
        hello = String()

    class Subscription(ObjectType):
        count_upwards = Field(Int, limit=Int(required=True))

        async def subscribe_count_upwards(root, info, limit):
            count = 0
            while count < limit:
                count += 1
                yield count

    schema = Schema(query=Query, subscription=Subscription)

    subscription = "subscription { countUpwards(limit: 5) }"
    result = await schema.subscribe(subscription)
    count = 0
    async for item in result:
        count = item.data["countUpwards"]
    assert count == 5



from ..resolver import (
    attr_resolver,
    dict_resolver,
    dict_or_attr_resolver,
    get_default_resolver,
    set_default_resolver,
)

args = {}
context = None
info = None

demo_dict = {"attr": "value"}


class demo_obj:
    attr = "value"


def test_attr_resolver():
    resolved = attr_resolver("attr", None, demo_obj, info, **args)
    assert resolved == "value"


def test_attr_resolver_default_value():
    resolved = attr_resolver("attr2", "default", demo_obj, info, **args)
    assert resolved == "default"


def test_dict_resolver():
    resolved = dict_resolver("attr", None, demo_dict, info, **args)
    assert resolved == "value"


def test_dict_resolver_default_value():
    resolved = dict_resolver("attr2", "default", demo_dict, info, **args)
    assert resolved == "default"


def test_dict_or_attr_resolver():
    resolved = dict_or_attr_resolver("attr", None, demo_dict, info, **args)
    assert resolved == "value"

    resolved = dict_or_attr_resolver("attr", None, demo_obj, info, **args)
    assert resolved == "value"


def test_get_default_resolver_is_attr_resolver():
    assert get_default_resolver() == dict_or_attr_resolver


def test_set_default_resolver_workd():
    default_resolver = get_default_resolver()

    set_default_resolver(dict_resolver)
    assert get_default_resolver() == dict_resolver

    set_default_resolver(default_resolver)



import re
from uuid import uuid4

from graphql import graphql_sync

from ..id_type import BaseGlobalIDType, SimpleGlobalIDType, UUIDGlobalIDType
from ..node import Node
from ...types import Int, ObjectType, Schema, String


class TestUUIDGlobalID:
    def setup_method(self):
        self.user_list = [
            {"id": uuid4(), "name": "First"},
            {"id": uuid4(), "name": "Second"},
            {"id": uuid4(), "name": "Third"},
            {"id": uuid4(), "name": "Fourth"},
        ]
        self.users = {user["id"]: user for user in self.user_list}

        class CustomNode(Node):
            class Meta:
                global_id_type = UUIDGlobalIDType

        class User(ObjectType):
            class Meta:
                interfaces = [CustomNode]

            name = String()

            @classmethod
            def get_node(cls, _type, _id):
                return self.users[_id]

        class RootQuery(ObjectType):
            user = CustomNode.Field(User)

        self.schema = Schema(query=RootQuery, types=[User])
        self.graphql_schema = self.schema.graphql_schema

    def test_str_schema_correct(self):
        """
        Check that the schema has the expected and custom node interface and user type and that they both use UUIDs
        """
        parsed = re.findall(r"(.+) \{\n\s*([\w\W]*?)\n\}", str(self.schema))
        types = [t for t, f in parsed]
        fields = [f for t, f in parsed]
        custom_node_interface = "interface CustomNode"
        assert custom_node_interface in types
        assert (
            '"""The ID of the object"""\n  id: UUID!'
            == fields[types.index(custom_node_interface)]
        )
        user_type = "type User implements CustomNode"
        assert user_type in types
        assert (
            '"""The ID of the object"""\n  id: UUID!\n  name: String'
            == fields[types.index(user_type)]
        )

    def test_get_by_id(self):
        query = """query userById($id: UUID!) {
            user(id: $id) {
                id
                name
            }
        }"""
        # UUID need to be converted to string for serialization
        result = graphql_sync(
            self.graphql_schema,
            query,
            variable_values={"id": str(self.user_list[0]["id"])},
        )
        assert not result.errors
        assert result.data["user"]["id"] == str(self.user_list[0]["id"])
        assert result.data["user"]["name"] == self.user_list[0]["name"]


class TestSimpleGlobalID:
    def setup_method(self):
        self.user_list = [
            {"id": "my global primary key in clear 1", "name": "First"},
            {"id": "my global primary key in clear 2", "name": "Second"},
            {"id": "my global primary key in clear 3", "name": "Third"},
            {"id": "my global primary key in clear 4", "name": "Fourth"},
        ]
        self.users = {user["id"]: user for user in self.user_list}

        class CustomNode(Node):
            class Meta:
                global_id_type = SimpleGlobalIDType

        class User(ObjectType):
            class Meta:
                interfaces = [CustomNode]

            name = String()

            @classmethod
            def get_node(cls, _type, _id):
                return self.users[_id]

        class RootQuery(ObjectType):
            user = CustomNode.Field(User)

        self.schema = Schema(query=RootQuery, types=[User])
        self.graphql_schema = self.schema.graphql_schema

    def test_str_schema_correct(self):
        """
        Check that the schema has the expected and custom node interface and user type and that they both use UUIDs
        """
        parsed = re.findall(r"(.+) \{\n\s*([\w\W]*?)\n\}", str(self.schema))
        types = [t for t, f in parsed]
        fields = [f for t, f in parsed]
        custom_node_interface = "interface CustomNode"
        assert custom_node_interface in types
        assert (
            '"""The ID of the object"""\n  id: ID!'
            == fields[types.index(custom_node_interface)]
        )
        user_type = "type User implements CustomNode"
        assert user_type in types
        assert (
            '"""The ID of the object"""\n  id: ID!\n  name: String'
            == fields[types.index(user_type)]
        )

    def test_get_by_id(self):
        query = """query {
            user(id: "my global primary key in clear 3") {
                id
                name
            }
        }"""
        result = graphql_sync(self.graphql_schema, query)
        assert not result.errors
        assert result.data["user"]["id"] == self.user_list[2]["id"]
        assert result.data["user"]["name"] == self.user_list[2]["name"]


class TestCustomGlobalID:
    def setup_method(self):
        self.user_list = [
            {"id": 1, "name": "First"},
            {"id": 2, "name": "Second"},
            {"id": 3, "name": "Third"},
            {"id": 4, "name": "Fourth"},
        ]
        self.users = {user["id"]: user for user in self.user_list}

        class CustomGlobalIDType(BaseGlobalIDType):
            """
            Global id that is simply and integer in clear.
            """

            graphene_type = Int

            @classmethod
            def resolve_global_id(cls, info, global_id):
                _type = info.return_type.graphene_type._meta.name
                return _type, global_id

            @classmethod
            def to_global_id(cls, _type, _id):
                return _id

        class CustomNode(Node):
            class Meta:
                global_id_type = CustomGlobalIDType

        class User(ObjectType):
            class Meta:
                interfaces = [CustomNode]

            name = String()

            @classmethod
            def get_node(cls, _type, _id):
                return self.users[_id]

        class RootQuery(ObjectType):
            user = CustomNode.Field(User)

        self.schema = Schema(query=RootQuery, types=[User])
        self.graphql_schema = self.schema.graphql_schema

    def test_str_schema_correct(self):
        """
        Check that the schema has the expected and custom node interface and user type and that they both use UUIDs
        """
        parsed = re.findall(r"(.+) \{\n\s*([\w\W]*?)\n\}", str(self.schema))
        types = [t for t, f in parsed]
        fields = [f for t, f in parsed]
        custom_node_interface = "interface CustomNode"
        assert custom_node_interface in types
        assert (
            '"""The ID of the object"""\n  id: Int!'
            == fields[types.index(custom_node_interface)]
        )
        user_type = "type User implements CustomNode"
        assert user_type in types
        assert (
            '"""The ID of the object"""\n  id: Int!\n  name: String'
            == fields[types.index(user_type)]
        )

    def test_get_by_id(self):
        query = """query {
            user(id: 2) {
                id
                name
            }
        }"""
        result = graphql_sync(self.graphql_schema, query)
        assert not result.errors
        assert result.data["user"]["id"] == self.user_list[1]["id"]
        assert result.data["user"]["name"] == self.user_list[1]["name"]


class TestIncompleteCustomGlobalID:
    def setup_method(self):
        self.user_list = [
            {"id": 1, "name": "First"},
            {"id": 2, "name": "Second"},
            {"id": 3, "name": "Third"},
            {"id": 4, "name": "Fourth"},
        ]
        self.users = {user["id"]: user for user in self.user_list}

    def test_must_define_to_global_id(self):
        """
        Test that if the `to_global_id` method is not defined, we can query the object, but we can't request its ID.
        """

        class CustomGlobalIDType(BaseGlobalIDType):
            graphene_type = Int

            @classmethod
            def resolve_global_id(cls, info, global_id):
                _type = info.return_type.graphene_type._meta.name
                return _type, global_id

        class CustomNode(Node):
            class Meta:
                global_id_type = CustomGlobalIDType

        class User(ObjectType):
            class Meta:
                interfaces = [CustomNode]

            name = String()

            @classmethod
            def get_node(cls, _type, _id):
                return self.users[_id]

        class RootQuery(ObjectType):
            user = CustomNode.Field(User)

        self.schema = Schema(query=RootQuery, types=[User])
        self.graphql_schema = self.schema.graphql_schema

        query = """query {
            user(id: 2) {
                name
            }
        }"""
        result = graphql_sync(self.graphql_schema, query)
        assert not result.errors
        assert result.data["user"]["name"] == self.user_list[1]["name"]

        query = """query {
            user(id: 2) {
                id
                name
            }
        }"""
        result = graphql_sync(self.graphql_schema, query)
        assert result.errors is not None
        assert len(result.errors) == 1
        assert result.errors[0].path == ["user", "id"]

    def test_must_define_resolve_global_id(self):
        """
        Test that if the `resolve_global_id` method is not defined, we can't query the object by ID.
        """

        class CustomGlobalIDType(BaseGlobalIDType):
            graphene_type = Int

            @classmethod
            def to_global_id(cls, _type, _id):
                return _id

        class CustomNode(Node):
            class Meta:
                global_id_type = CustomGlobalIDType

        class User(ObjectType):
            class Meta:
                interfaces = [CustomNode]

            name = String()

            @classmethod
            def get_node(cls, _type, _id):
                return self.users[_id]

        class RootQuery(ObjectType):
            user = CustomNode.Field(User)

        self.schema = Schema(query=RootQuery, types=[User])
        self.graphql_schema = self.schema.graphql_schema

        query = """query {
            user(id: 2) {
                id
                name
            }
        }"""
        result = graphql_sync(self.graphql_schema, query)
        assert result.errors is not None
        assert len(result.errors) == 1
        assert result.errors[0].path == ["user"]



from .. import deprecated
from ..resolve_only_args import resolve_only_args


def test_resolve_only_args(mocker):
    mocker.patch.object(deprecated, "warn_deprecation")

    def resolver(root, **args):
        return root, args

    wrapped_resolver = resolve_only_args(resolver)
    result = wrapped_resolver(1, 2, a=3)
    assert result == (1, {"a": 3})



from typing import TYPE_CHECKING

from .base import BaseOptions, BaseType
from .inputfield import InputField
from .unmountedtype import UnmountedType
from .utils import yank_fields_from_attrs

# For static type checking with type checker
if TYPE_CHECKING:
    from typing import Dict, Callable  # NOQA


class InputObjectTypeOptions(BaseOptions):
    fields = None  # type: Dict[str, InputField]
    container = None  # type: InputObjectTypeContainer


# Currently in Graphene, we get a `None` whenever we access an (optional) field that was not set in an InputObjectType
# using the InputObjectType.<attribute> dot access syntax. This is ambiguous, because in this current (Graphene
# historical) arrangement, we cannot distinguish between a field not being set and a field being set to None.
# At the same time, we shouldn't break existing code that expects a `None` when accessing a field that was not set.
_INPUT_OBJECT_TYPE_DEFAULT_VALUE = None

# To mitigate this, we provide the function `set_input_object_type_default_value` to allow users to change the default
# value returned in non-specified fields in InputObjectType to another meaningful sentinel value (e.g. Undefined)
# if they want to. This way, we can keep code that expects a `None` working while we figure out a better solution (or
# a well-documented breaking change) for this issue.


def set_input_object_type_default_value(default_value):
    """
    Change the sentinel value returned by non-specified fields in an InputObjectType
    Useful to differentiate between a field not being set and a field being set to None by using a sentinel value
    (e.g. Undefined is a good sentinel value for this purpose)

    This function should be called at the beginning of the app or in some other place where it is guaranteed to
    be called before any InputObjectType is defined.
    """
    global _INPUT_OBJECT_TYPE_DEFAULT_VALUE
    _INPUT_OBJECT_TYPE_DEFAULT_VALUE = default_value


class InputObjectTypeContainer(dict, BaseType):  # type: ignore
    class Meta:
        abstract = True

    def __init__(self, *args, **kwargs):
        dict.__init__(self, *args, **kwargs)
        for key in self._meta.fields:
            setattr(self, key, self.get(key, _INPUT_OBJECT_TYPE_DEFAULT_VALUE))

    def __init_subclass__(cls, *args, **kwargs):
        pass


class InputObjectType(UnmountedType, BaseType):
    """
    Input Object Type Definition

    An input object defines a structured collection of fields which may be
    supplied to a field argument.

    Using ``graphene.NonNull`` will ensure that a input value must be provided by the query.

    All class attributes of ``graphene.InputObjectType`` are implicitly mounted as InputField
    using the below Meta class options.

    .. code:: python

        from graphene import InputObjectType, String, InputField

        class Person(InputObjectType):
            # implicitly mounted as Input Field
            first_name = String(required=True)
            # explicitly mounted as Input Field
            last_name = InputField(String, description="Surname")

    The fields on an input object type can themselves refer to input object types, but you can't
    mix input and output types in your schema.

    Meta class options (optional):
        name (str): the name of the GraphQL type (must be unique in schema). Defaults to class
            name.
        description (str): the description of the GraphQL type in the schema. Defaults to class
            docstring.
        container (class): A class reference for a value object that allows for
            attribute initialization and access. Default InputObjectTypeContainer.
        fields (Dict[str, graphene.InputField]): Dictionary of field name to InputField. Not
            recommended to use (prefer class attributes).
    """

    @classmethod
    def __init_subclass_with_meta__(cls, container=None, _meta=None, **options):
        if not _meta:
            _meta = InputObjectTypeOptions(cls)

        fields = {}
        for base in reversed(cls.__mro__):
            fields.update(yank_fields_from_attrs(base.__dict__, _as=InputField))

        if _meta.fields:
            _meta.fields.update(fields)
        else:
            _meta.fields = fields
        if container is None:
            container = type(cls.__name__, (InputObjectTypeContainer, cls), {})
        _meta.container = container
        super(InputObjectType, cls).__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    def get_type(cls):
        """
        This function is called when the unmounted type (InputObjectType instance)
        is mounted (as a Field, InputField or Argument)
        """
        return cls



from ..objecttype import ObjectType, Field
from ..scalars import Scalar, Int, BigInt, Float, String, Boolean
from ..schema import Schema
from graphql import Undefined
from graphql.language.ast import IntValueNode


def test_scalar():
    class JSONScalar(Scalar):
        """Documentation"""

    assert JSONScalar._meta.name == "JSONScalar"
    assert JSONScalar._meta.description == "Documentation"


def test_ints():
    assert Int.parse_value(2**31 - 1) is not Undefined
    assert Int.parse_value("2.0") == 2
    assert Int.parse_value(2**31) is Undefined

    assert Int.parse_literal(IntValueNode(value=str(2**31 - 1))) == 2**31 - 1
    assert Int.parse_literal(IntValueNode(value=str(2**31))) is Undefined

    assert Int.parse_value(-(2**31)) is not Undefined
    assert Int.parse_value(-(2**31) - 1) is Undefined

    assert BigInt.parse_value(2**31) is not Undefined
    assert BigInt.parse_value("2.0") == 2
    assert BigInt.parse_value(-(2**31) - 1) is not Undefined

    assert BigInt.parse_literal(IntValueNode(value=str(2**31 - 1))) == 2**31 - 1
    assert BigInt.parse_literal(IntValueNode(value=str(2**31))) == 2**31


def return_input(_parent, _info, input):
    return input


class Optional(ObjectType):
    int = Int(input=Int(), resolver=return_input)
    big_int = BigInt(input=BigInt(), resolver=return_input)
    float = Float(input=Float(), resolver=return_input)
    bool = Boolean(input=Boolean(), resolver=return_input)
    string = String(input=String(), resolver=return_input)


class Query(ObjectType):
    optional = Field(Optional)

    def resolve_optional(self, info):
        return Optional()

    def resolve_required(self, info, input):
        return input


schema = Schema(query=Query)


class TestInt:
    def test_query(self):
        """
        Test that a normal query works.
        """
        result = schema.execute("{ optional { int(input: 20) } }")
        assert not result.errors
        assert result.data == {"optional": {"int": 20}}

    def test_optional_input(self):
        """
        Test that we can provide a null value to an optional input
        """
        result = schema.execute("{ optional { int(input: null) } }")
        assert not result.errors
        assert result.data == {"optional": {"int": None}}

    def test_invalid_input(self):
        """
        Test that if an invalid type is provided we get an error
        """
        result = schema.execute('{ optional { int(input: "20") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == 'Int cannot represent non-integer value: "20"'
        )

        result = schema.execute('{ optional { int(input: "a") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert result.errors[0].message == 'Int cannot represent non-integer value: "a"'

        result = schema.execute("{ optional { int(input: true) } }")
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == "Int cannot represent non-integer value: true"
        )


class TestBigInt:
    def test_query(self):
        """
        Test that a normal query works.
        """
        value = 2**31
        result = schema.execute("{ optional { bigInt(input: %s) } }" % value)
        assert not result.errors
        assert result.data == {"optional": {"bigInt": value}}

    def test_optional_input(self):
        """
        Test that we can provide a null value to an optional input
        """
        result = schema.execute("{ optional { bigInt(input: null) } }")
        assert not result.errors
        assert result.data == {"optional": {"bigInt": None}}

    def test_invalid_input(self):
        """
        Test that if an invalid type is provided we get an error
        """
        result = schema.execute('{ optional { bigInt(input: "20") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == "Expected value of type 'BigInt', found \"20\"."
        )

        result = schema.execute('{ optional { bigInt(input: "a") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == "Expected value of type 'BigInt', found \"a\"."
        )

        result = schema.execute("{ optional { bigInt(input: true) } }")
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == "Expected value of type 'BigInt', found true."
        )


class TestFloat:
    def test_query(self):
        """
        Test that a normal query works.
        """
        result = schema.execute("{ optional { float(input: 20) } }")
        assert not result.errors
        assert result.data == {"optional": {"float": 20.0}}

        result = schema.execute("{ optional { float(input: 20.2) } }")
        assert not result.errors
        assert result.data == {"optional": {"float": 20.2}}

    def test_optional_input(self):
        """
        Test that we can provide a null value to an optional input
        """
        result = schema.execute("{ optional { float(input: null) } }")
        assert not result.errors
        assert result.data == {"optional": {"float": None}}

    def test_invalid_input(self):
        """
        Test that if an invalid type is provided we get an error
        """
        result = schema.execute('{ optional { float(input: "20") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == 'Float cannot represent non numeric value: "20"'
        )

        result = schema.execute('{ optional { float(input: "a") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == 'Float cannot represent non numeric value: "a"'
        )

        result = schema.execute("{ optional { float(input: true) } }")
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == "Float cannot represent non numeric value: true"
        )


class TestBoolean:
    def test_query(self):
        """
        Test that a normal query works.
        """
        result = schema.execute("{ optional { bool(input: true) } }")
        assert not result.errors
        assert result.data == {"optional": {"bool": True}}

        result = schema.execute("{ optional { bool(input: false) } }")
        assert not result.errors
        assert result.data == {"optional": {"bool": False}}

    def test_optional_input(self):
        """
        Test that we can provide a null value to an optional input
        """
        result = schema.execute("{ optional { bool(input: null) } }")
        assert not result.errors
        assert result.data == {"optional": {"bool": None}}

    def test_invalid_input(self):
        """
        Test that if an invalid type is provided we get an error
        """
        result = schema.execute('{ optional { bool(input: "True") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message
            == 'Boolean cannot represent a non boolean value: "True"'
        )

        result = schema.execute('{ optional { bool(input: "true") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message
            == 'Boolean cannot represent a non boolean value: "true"'
        )

        result = schema.execute('{ optional { bool(input: "a") } }')
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message
            == 'Boolean cannot represent a non boolean value: "a"'
        )

        result = schema.execute("{ optional { bool(input: 1) } }")
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message
            == "Boolean cannot represent a non boolean value: 1"
        )

        result = schema.execute("{ optional { bool(input: 0) } }")
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message
            == "Boolean cannot represent a non boolean value: 0"
        )


class TestString:
    def test_query(self):
        """
        Test that a normal query works.
        """
        result = schema.execute('{ optional { string(input: "something something") } }')
        assert not result.errors
        assert result.data == {"optional": {"string": "something something"}}

        result = schema.execute('{ optional { string(input: "True") } }')
        assert not result.errors
        assert result.data == {"optional": {"string": "True"}}

        result = schema.execute('{ optional { string(input: "0") } }')
        assert not result.errors
        assert result.data == {"optional": {"string": "0"}}

    def test_optional_input(self):
        """
        Test that we can provide a null value to an optional input
        """
        result = schema.execute("{ optional { string(input: null) } }")
        assert not result.errors
        assert result.data == {"optional": {"string": None}}

    def test_invalid_input(self):
        """
        Test that if an invalid type is provided we get an error
        """
        result = schema.execute("{ optional { string(input: 1) } }")
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message == "String cannot represent a non string value: 1"
        )

        result = schema.execute("{ optional { string(input: 3.2) } }")
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message
            == "String cannot represent a non string value: 3.2"
        )

        result = schema.execute("{ optional { string(input: true) } }")
        assert result.errors
        assert len(result.errors) == 1
        assert (
            result.errors[0].message
            == "String cannot represent a non string value: true"
        )



def is_introspection_key(key):
    # from: https://spec.graphql.org/June2018/#sec-Schema
    # > All types and directives defined within a schema must not have a name which
    # > begins with "__" (two underscores), as this is used exclusively
    # > by GraphQL’s introspection system.
    return str(key).startswith("__")



from pytest import mark

from graphql_relay.utils import base64

from ...types import ObjectType, Schema, String
from ..connection import Connection, ConnectionField, PageInfo
from ..node import Node

letter_chars = ["A", "B", "C", "D", "E"]


class Letter(ObjectType):
    class Meta:
        interfaces = (Node,)

    letter = String()


class LetterConnection(Connection):
    class Meta:
        node = Letter


class Query(ObjectType):
    letters = ConnectionField(LetterConnection)
    connection_letters = ConnectionField(LetterConnection)
    async_letters = ConnectionField(LetterConnection)

    node = Node.Field()

    def resolve_letters(self, info, **args):
        return list(letters.values())

    async def resolve_async_letters(self, info, **args):
        return list(letters.values())

    def resolve_connection_letters(self, info, **args):
        return LetterConnection(
            page_info=PageInfo(has_next_page=True, has_previous_page=False),
            edges=[
                LetterConnection.Edge(node=Letter(id=0, letter="A"), cursor="a-cursor")
            ],
        )


schema = Schema(Query)

letters = {letter: Letter(id=i, letter=letter) for i, letter in enumerate(letter_chars)}


def edges(selected_letters):
    return [
        {
            "node": {"id": base64("Letter:%s" % letter.id), "letter": letter.letter},
            "cursor": base64("arrayconnection:%s" % letter.id),
        }
        for letter in [letters[i] for i in selected_letters]
    ]


def cursor_for(ltr):
    letter = letters[ltr]
    return base64("arrayconnection:%s" % letter.id)


async def execute(args=""):
    if args:
        args = "(" + args + ")"
    return await schema.execute_async(
        """
    {
        letters%s {
            edges {
                node {
                    id
                    letter
                }
                cursor
            }
            pageInfo {
                hasPreviousPage
                hasNextPage
                startCursor
                endCursor
            }
        }
    }
    """
        % args
    )


async def check(args, letters, has_previous_page=False, has_next_page=False):
    result = await execute(args)
    expected_edges = edges(letters)
    expected_page_info = {
        "hasPreviousPage": has_previous_page,
        "hasNextPage": has_next_page,
        "endCursor": expected_edges[-1]["cursor"] if expected_edges else None,
        "startCursor": expected_edges[0]["cursor"] if expected_edges else None,
    }

    assert not result.errors
    assert result.data == {
        "letters": {"edges": expected_edges, "pageInfo": expected_page_info}
    }


@mark.asyncio
async def test_returns_all_elements_without_filters():
    await check("", "ABCDE")


@mark.asyncio
async def test_respects_a_smaller_first():
    await check("first: 2", "AB", has_next_page=True)


@mark.asyncio
async def test_respects_an_overly_large_first():
    await check("first: 10", "ABCDE")


@mark.asyncio
async def test_respects_a_smaller_last():
    await check("last: 2", "DE", has_previous_page=True)


@mark.asyncio
async def test_respects_an_overly_large_last():
    await check("last: 10", "ABCDE")


@mark.asyncio
async def test_respects_first_and_after():
    await check(f'first: 2, after: "{cursor_for("B")}"', "CD", has_next_page=True)


@mark.asyncio
async def test_respects_first_and_after_with_long_first():
    await check(f'first: 10, after: "{cursor_for("B")}"', "CDE")


@mark.asyncio
async def test_respects_last_and_before():
    await check(f'last: 2, before: "{cursor_for("D")}"', "BC", has_previous_page=True)


@mark.asyncio
async def test_respects_last_and_before_with_long_last():
    await check(f'last: 10, before: "{cursor_for("D")}"', "ABC")


@mark.asyncio
async def test_respects_first_and_after_and_before_too_few():
    await check(
        f'first: 2, after: "{cursor_for("A")}", before: "{cursor_for("E")}"',
        "BC",
        has_next_page=True,
    )


@mark.asyncio
async def test_respects_first_and_after_and_before_too_many():
    await check(
        f'first: 4, after: "{cursor_for("A")}", before: "{cursor_for("E")}"', "BCD"
    )


@mark.asyncio
async def test_respects_first_and_after_and_before_exactly_right():
    await check(
        f'first: 3, after: "{cursor_for("A")}", before: "{cursor_for("E")}"', "BCD"
    )


@mark.asyncio
async def test_respects_last_and_after_and_before_too_few():
    await check(
        f'last: 2, after: "{cursor_for("A")}", before: "{cursor_for("E")}"',
        "CD",
        has_previous_page=True,
    )


@mark.asyncio
async def test_respects_last_and_after_and_before_too_many():
    await check(
        f'last: 4, after: "{cursor_for("A")}", before: "{cursor_for("E")}"', "BCD"
    )


@mark.asyncio
async def test_respects_last_and_after_and_before_exactly_right():
    await check(
        f'last: 3, after: "{cursor_for("A")}", before: "{cursor_for("E")}"', "BCD"
    )


@mark.asyncio
async def test_returns_no_elements_if_first_is_0():
    await check("first: 0", "", has_next_page=True)


@mark.asyncio
async def test_returns_all_elements_if_cursors_are_invalid():
    await check('before: "invalid" after: "invalid"', "ABCDE")


@mark.asyncio
async def test_returns_all_elements_if_cursors_are_on_the_outside():
    await check(
        f'before: "{base64("arrayconnection:%s" % 6)}" after: "{base64("arrayconnection:%s" % -1)}"',
        "ABCDE",
    )


@mark.asyncio
async def test_returns_no_elements_if_cursors_cross():
    await check(
        f'before: "{base64("arrayconnection:%s" % 2)}" after: "{base64("arrayconnection:%s" % 4)}"',
        "",
    )


@mark.asyncio
async def test_connection_type_nodes():
    result = await schema.execute_async(
        """
    {
        connectionLetters {
            edges {
                node {
                    id
                    letter
                }
                cursor
            }
            pageInfo {
                hasPreviousPage
                hasNextPage
            }
        }
    }
    """
    )

    assert not result.errors
    assert result.data == {
        "connectionLetters": {
            "edges": [
                {"node": {"id": "TGV0dGVyOjA=", "letter": "A"}, "cursor": "a-cursor"}
            ],
            "pageInfo": {"hasPreviousPage": False, "hasNextPage": True},
        }
    }


@mark.asyncio
async def test_connection_async():
    result = await schema.execute_async(
        """
    {
        asyncLetters(first:1) {
            edges {
                node {
                    id
                    letter
                }
            }
            pageInfo {
                hasPreviousPage
                hasNextPage
            }
        }
    }
    """
    )

    assert not result.errors
    assert result.data == {
        "asyncLetters": {
            "edges": [{"node": {"id": "TGV0dGVyOjA=", "letter": "A"}}],
            "pageInfo": {"hasPreviousPage": False, "hasNextPage": True},
        }
    }



import inspect
from collections.abc import Mapping
from functools import partial

from .argument import Argument, to_arguments
from .mountedtype import MountedType
from .resolver import default_resolver
from .structures import NonNull
from .unmountedtype import UnmountedType
from .utils import get_type
from ..utils.deprecated import warn_deprecation

base_type = type


def source_resolver(source, root, info, **args):
    resolved = default_resolver(source, None, root, info, **args)
    if inspect.isfunction(resolved) or inspect.ismethod(resolved):
        return resolved()
    return resolved


class Field(MountedType):
    """
    Makes a field available on an ObjectType in the GraphQL schema. Any type can be mounted as a
    Field:

    - Object Type
    - Scalar Type
    - Enum
    - Interface
    - Union

    All class attributes of ``graphene.ObjectType`` are implicitly mounted as Field using the below
    arguments.

    .. code:: python

        class Person(ObjectType):
            first_name = graphene.String(required=True)                # implicitly mounted as Field
            last_name = graphene.Field(String, description='Surname')  # explicitly mounted as Field

    args:
        type (class for a graphene.UnmountedType): Must be a class (not an instance) of an
            unmounted graphene type (ex. scalar or object) which is used for the type of this
            field in the GraphQL schema. You can provide a dotted module import path (string)
            to the class instead of the class itself (e.g. to avoid circular import issues).
        args (optional, Dict[str, graphene.Argument]): Arguments that can be input to the field.
            Prefer to use ``**extra_args``, unless you use an argument name that clashes with one
            of the Field arguments presented here (see :ref:`example<ResolverParamGraphQLArguments>`).
        resolver (optional, Callable): A function to get the value for a Field from the parent
            value object. If not set, the default resolver method for the schema is used.
        source (optional, str): attribute name to resolve for this field from the parent value
            object. Alternative to resolver (cannot set both source and resolver).
        deprecation_reason (optional, str): Setting this value indicates that the field is
            depreciated and may provide instruction or reason on how for clients to proceed.
        required (optional, bool): indicates this field as not null in the graphql schema. Same behavior as
            graphene.NonNull. Default False.
        name (optional, str): the name of the GraphQL field (must be unique in a type). Defaults to attribute
            name.
        description (optional, str): the description of the GraphQL field in the schema.
        default_value (optional, Any): Default value to resolve if none set from schema.
        **extra_args (optional, Dict[str, Union[graphene.Argument, graphene.UnmountedType]): any
            additional arguments to mount on the field.
    """

    def __init__(
        self,
        type_,
        args=None,
        resolver=None,
        source=None,
        deprecation_reason=None,
        name=None,
        description=None,
        required=False,
        _creation_counter=None,
        default_value=None,
        **extra_args,
    ):
        super(Field, self).__init__(_creation_counter=_creation_counter)
        assert not args or isinstance(
            args, Mapping
        ), f'Arguments in a field have to be a mapping, received "{args}".'
        assert not (
            source and resolver
        ), "A Field cannot have a source and a resolver in at the same time."
        assert not callable(
            default_value
        ), f'The default value can not be a function but received "{base_type(default_value)}".'

        if required:
            type_ = NonNull(type_)

        # Check if name is actually an argument of the field
        if isinstance(name, (Argument, UnmountedType)):
            extra_args["name"] = name
            name = None

        # Check if source is actually an argument of the field
        if isinstance(source, (Argument, UnmountedType)):
            extra_args["source"] = source
            source = None

        self.name = name
        self._type = type_
        self.args = to_arguments(args or {}, extra_args)
        if source:
            resolver = partial(source_resolver, source)
        self.resolver = resolver
        self.deprecation_reason = deprecation_reason
        self.description = description
        self.default_value = default_value

    @property
    def type(self):
        return get_type(self._type)

    get_resolver = None

    def wrap_resolve(self, parent_resolver):
        """
        Wraps a function resolver, using the ObjectType resolve_{FIELD_NAME}
        (parent_resolver) if the Field definition has no resolver.
        """
        if self.get_resolver is not None:
            warn_deprecation(
                "The get_resolver method is being deprecated, please rename it to wrap_resolve."
            )
            return self.get_resolver(parent_resolver)

        return self.resolver or parent_resolver

    def wrap_subscribe(self, parent_subscribe):
        """
        Wraps a function subscribe, using the ObjectType subscribe_{FIELD_NAME}
        (parent_subscribe) if the Field definition has no subscribe.
        """
        return parent_subscribe



from pytest import raises

from ..argument import Argument
from ..dynamic import Dynamic
from ..mutation import Mutation
from ..objecttype import ObjectType
from ..scalars import String
from ..schema import Schema
from ..structures import NonNull
from ..interface import Interface


class MyType(Interface):
    pass


def test_generate_mutation_no_args():
    class MyMutation(Mutation):
        """Documentation"""

        def mutate(self, info, **args):
            return args

    assert issubclass(MyMutation, ObjectType)
    assert MyMutation._meta.name == "MyMutation"
    assert MyMutation._meta.description == "Documentation"
    resolved = MyMutation.Field().resolver(None, None, name="Peter")
    assert resolved == {"name": "Peter"}


def test_generate_mutation_with_meta():
    class MyMutation(Mutation):
        class Meta:
            name = "MyOtherMutation"
            description = "Documentation"
            interfaces = (MyType,)

        def mutate(self, info, **args):
            return args

    assert MyMutation._meta.name == "MyOtherMutation"
    assert MyMutation._meta.description == "Documentation"
    assert MyMutation._meta.interfaces == (MyType,)
    resolved = MyMutation.Field().resolver(None, None, name="Peter")
    assert resolved == {"name": "Peter"}


def test_mutation_raises_exception_if_no_mutate():
    with raises(AssertionError) as excinfo:

        class MyMutation(Mutation):
            pass

    assert "All mutations must define a mutate method in it" == str(excinfo.value)


def test_mutation_custom_output_type():
    class User(ObjectType):
        name = String()

    class CreateUser(Mutation):
        class Arguments:
            name = String()

        Output = User

        def mutate(self, info, name):
            return User(name=name)

    field = CreateUser.Field()
    assert field.type == User
    assert field.args == {"name": Argument(String)}
    resolved = field.resolver(None, None, name="Peter")
    assert isinstance(resolved, User)
    assert resolved.name == "Peter"


def test_mutation_execution():
    class CreateUser(Mutation):
        class Arguments:
            name = String()
            dynamic = Dynamic(lambda: String())
            dynamic_none = Dynamic(lambda: None)

        name = String()
        dynamic = Dynamic(lambda: String())

        def mutate(self, info, name, dynamic):
            return CreateUser(name=name, dynamic=dynamic)

    class Query(ObjectType):
        a = String()

    class MyMutation(ObjectType):
        create_user = CreateUser.Field()

    schema = Schema(query=Query, mutation=MyMutation)
    result = schema.execute(
        """ mutation mymutation {
        createUser(name:"Peter", dynamic: "dynamic") {
            name
            dynamic
        }
    }
    """
    )
    assert not result.errors
    assert result.data == {"createUser": {"name": "Peter", "dynamic": "dynamic"}}


def test_mutation_no_fields_output():
    class CreateUser(Mutation):
        name = String()

        def mutate(self, info):
            return CreateUser()

    class Query(ObjectType):
        a = String()

    class MyMutation(ObjectType):
        create_user = CreateUser.Field()

    schema = Schema(query=Query, mutation=MyMutation)
    result = schema.execute(
        """ mutation mymutation {
        createUser {
            name
        }
    }
    """
    )
    assert not result.errors
    assert result.data == {"createUser": {"name": None}}


def test_mutation_allow_to_have_custom_args():
    class CreateUser(Mutation):
        class Arguments:
            name = String()

        name = String()

        def mutate(self, info, name):
            return CreateUser(name=name)

    class MyMutation(ObjectType):
        create_user = CreateUser.Field(
            name="createUser",
            description="Create a user",
            deprecation_reason="Is deprecated",
            required=True,
        )

    field = MyMutation._meta.fields["create_user"]
    assert field.name == "createUser"
    assert field.description == "Create a user"
    assert field.deprecation_reason == "Is deprecated"
    assert field.type == NonNull(CreateUser)


def test_mutation_default_args_output():
    class CreateUser(Mutation):
        """Description."""

        class Arguments:
            name = String()

        name = String()

        def mutate(self, info, name):
            return CreateUser(name=name)

    class MyMutation(ObjectType):
        create_user = CreateUser.Field()

    field = MyMutation._meta.fields["create_user"]
    assert field.name is None
    assert field.description == "Description."
    assert field.deprecation_reason is None
    assert field.type == CreateUser


def test_mutation_as_subclass():
    class BaseCreateUser(Mutation):
        class Arguments:
            name = String()

        name = String()

        def mutate(self, info, **args):
            return args

    class CreateUserWithPlanet(BaseCreateUser):
        class Arguments(BaseCreateUser.Arguments):
            planet = String()

        planet = String()

        def mutate(self, info, **args):
            return CreateUserWithPlanet(**args)

    class MyMutation(ObjectType):
        create_user_with_planet = CreateUserWithPlanet.Field()

    class Query(ObjectType):
        a = String()

    schema = Schema(query=Query, mutation=MyMutation)
    result = schema.execute(
        """ mutation mymutation {
        createUserWithPlanet(name:"Peter", planet: "earth") {
            name
            planet
        }
    }
    """
    )
    assert not result.errors
    assert result.data == {"createUserWithPlanet": {"name": "Peter", "planet": "earth"}}



import re

from pytest import raises

from ...types import Argument, Field, Int, List, NonNull, ObjectType, Schema, String
from ..connection import (
    Connection,
    ConnectionField,
    PageInfo,
    ConnectionOptions,
    get_edge_class,
)
from ..node import Node


class MyObject(ObjectType):
    class Meta:
        interfaces = [Node]

    field = String()


def test_connection():
    class MyObjectConnection(Connection):
        extra = String()

        class Meta:
            node = MyObject

        class Edge:
            other = String()

    assert MyObjectConnection._meta.name == "MyObjectConnection"
    fields = MyObjectConnection._meta.fields
    assert list(fields) == ["page_info", "edges", "extra"]
    edge_field = fields["edges"]
    pageinfo_field = fields["page_info"]

    assert isinstance(edge_field, Field)
    assert isinstance(edge_field.type, NonNull)
    assert isinstance(edge_field.type.of_type, List)
    assert edge_field.type.of_type.of_type == MyObjectConnection.Edge

    assert isinstance(pageinfo_field, Field)
    assert isinstance(pageinfo_field.type, NonNull)
    assert pageinfo_field.type.of_type == PageInfo


def test_connection_inherit_abstracttype():
    class BaseConnection:
        extra = String()

    class MyObjectConnection(BaseConnection, Connection):
        class Meta:
            node = MyObject

    assert MyObjectConnection._meta.name == "MyObjectConnection"
    fields = MyObjectConnection._meta.fields
    assert list(fields) == ["page_info", "edges", "extra"]


def test_connection_extra_abstract_fields():
    class ConnectionWithNodes(Connection):
        class Meta:
            abstract = True

        @classmethod
        def __init_subclass_with_meta__(cls, node=None, name=None, **options):
            _meta = ConnectionOptions(cls)

            _meta.fields = {
                "nodes": Field(
                    NonNull(List(node)),
                    description="Contains all the nodes in this connection.",
                ),
            }

            return super(ConnectionWithNodes, cls).__init_subclass_with_meta__(
                node=node, name=name, _meta=_meta, **options
            )

    class MyObjectConnection(ConnectionWithNodes):
        class Meta:
            node = MyObject

        class Edge:
            other = String()

    assert MyObjectConnection._meta.name == "MyObjectConnection"
    fields = MyObjectConnection._meta.fields
    assert list(fields) == ["nodes", "page_info", "edges"]
    edge_field = fields["edges"]
    pageinfo_field = fields["page_info"]
    nodes_field = fields["nodes"]

    assert isinstance(edge_field, Field)
    assert isinstance(edge_field.type, NonNull)
    assert isinstance(edge_field.type.of_type, List)
    assert edge_field.type.of_type.of_type == MyObjectConnection.Edge

    assert isinstance(pageinfo_field, Field)
    assert isinstance(pageinfo_field.type, NonNull)
    assert pageinfo_field.type.of_type == PageInfo

    assert isinstance(nodes_field, Field)
    assert isinstance(nodes_field.type, NonNull)
    assert isinstance(nodes_field.type.of_type, List)
    assert nodes_field.type.of_type.of_type == MyObject


def test_connection_override_fields():
    class ConnectionWithNodes(Connection):
        class Meta:
            abstract = True

        @classmethod
        def __init_subclass_with_meta__(cls, node=None, name=None, **options):
            _meta = ConnectionOptions(cls)
            base_name = (
                re.sub("Connection$", "", name or cls.__name__) or node._meta.name
            )

            edge_class = get_edge_class(cls, node, base_name)

            _meta.fields = {
                "page_info": Field(
                    NonNull(
                        PageInfo,
                        name="pageInfo",
                        required=True,
                        description="Pagination data for this connection.",
                    )
                ),
                "edges": Field(
                    NonNull(List(NonNull(edge_class))),
                    description="Contains the nodes in this connection.",
                ),
            }

            return super(ConnectionWithNodes, cls).__init_subclass_with_meta__(
                node=node, name=name, _meta=_meta, **options
            )

    class MyObjectConnection(ConnectionWithNodes):
        class Meta:
            node = MyObject

    assert MyObjectConnection._meta.name == "MyObjectConnection"
    fields = MyObjectConnection._meta.fields
    assert list(fields) == ["page_info", "edges"]
    edge_field = fields["edges"]
    pageinfo_field = fields["page_info"]

    assert isinstance(edge_field, Field)
    assert isinstance(edge_field.type, NonNull)
    assert isinstance(edge_field.type.of_type, List)
    assert isinstance(edge_field.type.of_type.of_type, NonNull)

    assert edge_field.type.of_type.of_type.of_type.__name__ == "MyObjectEdge"

    # This page info is NonNull
    assert isinstance(pageinfo_field, Field)
    assert isinstance(edge_field.type, NonNull)
    assert pageinfo_field.type.of_type == PageInfo


def test_connection_name():
    custom_name = "MyObjectCustomNameConnection"

    class BaseConnection:
        extra = String()

    class MyObjectConnection(BaseConnection, Connection):
        class Meta:
            node = MyObject
            name = custom_name

    assert MyObjectConnection._meta.name == custom_name


def test_edge():
    class MyObjectConnection(Connection):
        class Meta:
            node = MyObject

        class Edge:
            other = String()

    Edge = MyObjectConnection.Edge
    assert Edge._meta.name == "MyObjectEdge"
    edge_fields = Edge._meta.fields
    assert list(edge_fields) == ["node", "cursor", "other"]

    assert isinstance(edge_fields["node"], Field)
    assert edge_fields["node"].type == MyObject

    assert isinstance(edge_fields["other"], Field)
    assert edge_fields["other"].type == String


def test_edge_with_bases():
    class BaseEdge:
        extra = String()

    class MyObjectConnection(Connection):
        class Meta:
            node = MyObject

        class Edge(BaseEdge):
            other = String()

    Edge = MyObjectConnection.Edge
    assert Edge._meta.name == "MyObjectEdge"
    edge_fields = Edge._meta.fields
    assert list(edge_fields) == ["node", "cursor", "extra", "other"]

    assert isinstance(edge_fields["node"], Field)
    assert edge_fields["node"].type == MyObject

    assert isinstance(edge_fields["other"], Field)
    assert edge_fields["other"].type == String


def test_edge_with_nonnull_node():
    class MyObjectConnection(Connection):
        class Meta:
            node = NonNull(MyObject)

    edge_fields = MyObjectConnection.Edge._meta.fields
    assert isinstance(edge_fields["node"], Field)
    assert isinstance(edge_fields["node"].type, NonNull)
    assert edge_fields["node"].type.of_type == MyObject


def test_pageinfo():
    assert PageInfo._meta.name == "PageInfo"
    fields = PageInfo._meta.fields
    assert list(fields) == [
        "has_next_page",
        "has_previous_page",
        "start_cursor",
        "end_cursor",
    ]


def test_connectionfield():
    class MyObjectConnection(Connection):
        class Meta:
            node = MyObject

    field = ConnectionField(MyObjectConnection)
    assert field.args == {
        "before": Argument(String),
        "after": Argument(String),
        "first": Argument(Int),
        "last": Argument(Int),
    }


def test_connectionfield_node_deprecated():
    field = ConnectionField(MyObject)
    with raises(Exception) as exc_info:
        field.type

    assert "ConnectionFields now need a explicit ConnectionType for Nodes." in str(
        exc_info.value
    )


def test_connectionfield_custom_args():
    class MyObjectConnection(Connection):
        class Meta:
            node = MyObject

    field = ConnectionField(
        MyObjectConnection, before=String(required=True), extra=String()
    )
    assert field.args == {
        "before": Argument(NonNull(String)),
        "after": Argument(String),
        "first": Argument(Int),
        "last": Argument(Int),
        "extra": Argument(String),
    }


def test_connectionfield_required():
    class MyObjectConnection(Connection):
        class Meta:
            node = MyObject

    class Query(ObjectType):
        test_connection = ConnectionField(MyObjectConnection, required=True)

        def resolve_test_connection(root, info, **args):
            return []

    schema = Schema(query=Query)
    executed = schema.execute("{ testConnection { edges { cursor } } }")
    assert not executed.errors
    assert executed.data == {"testConnection": {"edges": []}}


def test_connectionfield_strict_types():
    class MyObjectConnection(Connection):
        class Meta:
            node = MyObject
            strict_types = True

    connection_field = ConnectionField(MyObjectConnection)
    edges_field_type = connection_field.type._meta.fields["edges"].type
    assert isinstance(edges_field_type, NonNull)

    edges_list_element_type = edges_field_type.of_type.of_type
    assert isinstance(edges_list_element_type, NonNull)

    node_field = edges_list_element_type.of_type._meta.fields["node"]
    assert isinstance(node_field.type, NonNull)



import base64

from graphql import GraphQLError

from ..objecttype import ObjectType
from ..scalars import String
from ..schema import Schema
from ..base64 import Base64


class Query(ObjectType):
    base64 = Base64(_in=Base64(name="input"), _match=String(name="match"))
    bytes_as_base64 = Base64()
    string_as_base64 = Base64()
    number_as_base64 = Base64()

    def resolve_base64(self, info, _in=None, _match=None):
        if _match:
            assert _in == _match
        return _in

    def resolve_bytes_as_base64(self, info):
        return b"Hello world"

    def resolve_string_as_base64(self, info):
        return "Spam and eggs"

    def resolve_number_as_base64(self, info):
        return 42


schema = Schema(query=Query)


def test_base64_query():
    base64_value = base64.b64encode(b"Random string").decode("utf-8")
    result = schema.execute(
        """{{ base64(input: "{}", match: "Random string") }}""".format(base64_value)
    )
    assert not result.errors
    assert result.data == {"base64": base64_value}


def test_base64_query_with_variable():
    base64_value = base64.b64encode(b"Another string").decode("utf-8")

    # test datetime variable in string representation
    result = schema.execute(
        """
        query GetBase64($base64: Base64) {
            base64(input: $base64, match: "Another string")
        }
        """,
        variables={"base64": base64_value},
    )
    assert not result.errors
    assert result.data == {"base64": base64_value}


def test_base64_query_none():
    result = schema.execute("""{ base64 }""")
    assert not result.errors
    assert result.data == {"base64": None}


def test_base64_query_invalid():
    bad_inputs = [dict(), 123, "This is not valid base64"]

    for input_ in bad_inputs:
        result = schema.execute(
            """{ base64(input: $input) }""", variables={"input": input_}
        )
        assert isinstance(result.errors, list)
        assert len(result.errors) == 1
        assert isinstance(result.errors[0], GraphQLError)
        assert result.data is None


def test_base64_from_bytes():
    base64_value = base64.b64encode(b"Hello world").decode("utf-8")
    result = schema.execute("""{ bytesAsBase64 }""")
    assert not result.errors
    assert result.data == {"bytesAsBase64": base64_value}


def test_base64_from_string():
    base64_value = base64.b64encode(b"Spam and eggs").decode("utf-8")
    result = schema.execute("""{ stringAsBase64 }""")
    assert not result.errors
    assert result.data == {"stringAsBase64": base64_value}


def test_base64_from_number():
    base64_value = base64.b64encode(b"42").decode("utf-8")
    result = schema.execute("""{ numberAsBase64 }""")
    assert not result.errors
    assert result.data == {"numberAsBase64": base64_value}



import datetime

from graphql import GraphQLError

from pytest import fixture

from ..datetime import Date, DateTime, Time
from ..objecttype import ObjectType
from ..schema import Schema


class Query(ObjectType):
    datetime = DateTime(_in=DateTime(name="in"))
    date = Date(_in=Date(name="in"))
    time = Time(_at=Time(name="at"))

    def resolve_datetime(self, info, _in=None):
        return _in

    def resolve_date(self, info, _in=None):
        return _in

    def resolve_time(self, info, _at=None):
        return _at


schema = Schema(query=Query)


@fixture
def sample_datetime():
    utc_datetime = datetime.datetime(2019, 5, 25, 5, 30, 15, 10, datetime.timezone.utc)
    return utc_datetime


@fixture
def sample_time(sample_datetime):
    time = datetime.time(
        sample_datetime.hour,
        sample_datetime.minute,
        sample_datetime.second,
        sample_datetime.microsecond,
        sample_datetime.tzinfo,
    )
    return time


@fixture
def sample_date(sample_datetime):
    date = sample_datetime.date()
    return date


def test_datetime_query(sample_datetime):
    isoformat = sample_datetime.isoformat()

    result = schema.execute("""{ datetime(in: "%s") }""" % isoformat)
    assert not result.errors
    assert result.data == {"datetime": isoformat}


def test_datetime_query_with_variables(sample_datetime):
    isoformat = sample_datetime.isoformat()

    result = schema.execute(
        """
        query GetDate($datetime: DateTime) {
          literal: datetime(in: "%s")
          value: datetime(in: $datetime)
        }
        """
        % isoformat,
        variable_values={"datetime": isoformat},
    )
    assert not result.errors
    assert result.data == {"literal": isoformat, "value": isoformat}


def test_date_query(sample_date):
    isoformat = sample_date.isoformat()

    result = schema.execute("""{ date(in: "%s") }""" % isoformat)
    assert not result.errors
    assert result.data == {"date": isoformat}


def test_date_query_with_variables(sample_date):
    isoformat = sample_date.isoformat()

    result = schema.execute(
        """
        query GetDate($date: Date) {
          literal: date(in: "%s")
          value: date(in: $date)
        }
        """
        % isoformat,
        variable_values={"date": isoformat},
    )
    assert not result.errors
    assert result.data == {"literal": isoformat, "value": isoformat}


def test_time_query(sample_time):
    isoformat = sample_time.isoformat()

    result = schema.execute("""{ time(at: "%s") }""" % isoformat)
    assert not result.errors
    assert result.data == {"time": isoformat}


def test_time_query_with_variables(sample_time):
    isoformat = sample_time.isoformat()

    result = schema.execute(
        """
        query GetTime($time: Time) {
          literal: time(at: "%s")
          value: time(at: $time)
        }
        """
        % isoformat,
        variable_values={"time": isoformat},
    )
    assert not result.errors
    assert result.data == {"literal": isoformat, "value": isoformat}


def test_bad_datetime_query():
    not_a_date = "Some string that's not a datetime"

    result = schema.execute("""{ datetime(in: "%s") }""" % not_a_date)

    assert result.errors and len(result.errors) == 1
    error = result.errors[0]
    assert isinstance(error, GraphQLError)
    assert (
        error.message == "DateTime cannot represent value:"
        ' "Some string that\'s not a datetime"'
    )
    assert result.data is None


def test_bad_date_query():
    not_a_date = "Some string that's not a date"

    result = schema.execute("""{ date(in: "%s") }""" % not_a_date)

    error = result.errors[0]
    assert isinstance(error, GraphQLError)
    assert (
        error.message == "Date cannot represent value:"
        ' "Some string that\'s not a date"'
    )
    assert result.data is None


def test_bad_time_query():
    not_a_date = "Some string that's not a time"

    result = schema.execute("""{ time(at: "%s") }""" % not_a_date)

    error = result.errors[0]
    assert isinstance(error, GraphQLError)
    assert (
        error.message == "Time cannot represent value:"
        ' "Some string that\'s not a time"'
    )
    assert result.data is None


def test_datetime_query_variable(sample_datetime):
    isoformat = sample_datetime.isoformat()

    # test datetime variable provided as Python datetime
    result = schema.execute(
        """query Test($date: DateTime){ datetime(in: $date) }""",
        variables={"date": sample_datetime},
    )
    assert not result.errors
    assert result.data == {"datetime": isoformat}

    # test datetime variable in string representation
    result = schema.execute(
        """query Test($date: DateTime){ datetime(in: $date) }""",
        variables={"date": isoformat},
    )
    assert not result.errors
    assert result.data == {"datetime": isoformat}


def test_date_query_variable(sample_date):
    isoformat = sample_date.isoformat()

    # test date variable provided as Python date
    result = schema.execute(
        """query Test($date: Date){ date(in: $date) }""",
        variables={"date": sample_date},
    )
    assert not result.errors
    assert result.data == {"date": isoformat}

    # test date variable in string representation
    result = schema.execute(
        """query Test($date: Date){ date(in: $date) }""", variables={"date": isoformat}
    )
    assert not result.errors
    assert result.data == {"date": isoformat}


def test_time_query_variable(sample_time):
    isoformat = sample_time.isoformat()

    # test time variable provided as Python time
    result = schema.execute(
        """query Test($time: Time){ time(at: $time) }""",
        variables={"time": sample_time},
    )
    assert not result.errors
    assert result.data == {"time": isoformat}

    # test time variable in string representation
    result = schema.execute(
        """query Test($time: Time){ time(at: $time) }""", variables={"time": isoformat}
    )
    assert not result.errors
    assert result.data == {"time": isoformat}


def test_support_isoformat():
    isoformat = "2011-11-04T00:05:23Z"

    # test time variable provided as Python time
    result = schema.execute(
        """query DateTime($time: DateTime){ datetime(in: $time) }""",
        variables={"time": isoformat},
    )
    assert not result.errors
    assert result.data == {"datetime": "2011-11-04T00:05:23+00:00"}


def test_bad_variables(sample_date, sample_datetime, sample_time):
    def _test_bad_variables(type_, input_):
        result = schema.execute(
            f"""query Test($input: {type_}){{ {type_.lower()}(in: $input) }}""",
            variables={"input": input_},
        )
        assert isinstance(result.errors, list)
        assert len(result.errors) == 1
        assert isinstance(result.errors[0], GraphQLError)
        assert result.data is None

    not_a_date = dict()
    not_a_date_str = "Some string that's not a date"
    today = sample_date
    now = sample_datetime
    time = sample_time

    bad_pairs = [
        ("DateTime", not_a_date),
        ("DateTime", not_a_date_str),
        ("DateTime", today),
        ("DateTime", time),
        ("Date", not_a_date),
        ("Date", not_a_date_str),
        ("Date", time),
        ("Time", not_a_date),
        ("Time", not_a_date_str),
        ("Time", now),
        ("Time", today),
    ]

    for type_, input_ in bad_pairs:
        _test_bad_variables(type_, input_)



import json
from functools import partial

from graphql import (
    GraphQLError,
    GraphQLResolveInfo as ResolveInfo,
    Source,
    execute,
    parse,
)

from ..context import Context
from ..dynamic import Dynamic
from ..field import Field
from ..inputfield import InputField
from ..inputobjecttype import InputObjectType
from ..interface import Interface
from ..objecttype import ObjectType
from ..scalars import Boolean, Int, String
from ..schema import Schema
from ..structures import List, NonNull
from ..union import Union


def test_query():
    class Query(ObjectType):
        hello = String(resolver=lambda *_: "World")

    hello_schema = Schema(Query)

    executed = hello_schema.execute("{ hello }")
    assert not executed.errors
    assert executed.data == {"hello": "World"}


def test_query_source():
    class Root:
        _hello = "World"

        def hello(self):
            return self._hello

    class Query(ObjectType):
        hello = String(source="hello")

    hello_schema = Schema(Query)

    executed = hello_schema.execute("{ hello }", Root())
    assert not executed.errors
    assert executed.data == {"hello": "World"}


def test_query_union():
    class one_object:
        pass

    class two_object:
        pass

    class One(ObjectType):
        one = String()

        @classmethod
        def is_type_of(cls, root, info):
            return isinstance(root, one_object)

    class Two(ObjectType):
        two = String()

        @classmethod
        def is_type_of(cls, root, info):
            return isinstance(root, two_object)

    class MyUnion(Union):
        class Meta:
            types = (One, Two)

    class Query(ObjectType):
        unions = List(MyUnion)

        def resolve_unions(self, info):
            return [one_object(), two_object()]

    hello_schema = Schema(Query)

    executed = hello_schema.execute("{ unions { __typename } }")
    assert not executed.errors
    assert executed.data == {"unions": [{"__typename": "One"}, {"__typename": "Two"}]}


def test_query_interface():
    class one_object:
        pass

    class two_object:
        pass

    class MyInterface(Interface):
        base = String()

    class One(ObjectType):
        class Meta:
            interfaces = (MyInterface,)

        one = String()

        @classmethod
        def is_type_of(cls, root, info):
            return isinstance(root, one_object)

    class Two(ObjectType):
        class Meta:
            interfaces = (MyInterface,)

        two = String()

        @classmethod
        def is_type_of(cls, root, info):
            return isinstance(root, two_object)

    class Query(ObjectType):
        interfaces = List(MyInterface)

        def resolve_interfaces(self, info):
            return [one_object(), two_object()]

    hello_schema = Schema(Query, types=[One, Two])

    executed = hello_schema.execute("{ interfaces { __typename } }")
    assert not executed.errors
    assert executed.data == {
        "interfaces": [{"__typename": "One"}, {"__typename": "Two"}]
    }


def test_query_dynamic():
    class Query(ObjectType):
        hello = Dynamic(lambda: String(resolver=lambda *_: "World"))
        hellos = Dynamic(lambda: List(String, resolver=lambda *_: ["Worlds"]))
        hello_field = Dynamic(lambda: Field(String, resolver=lambda *_: "Field World"))

    hello_schema = Schema(Query)

    executed = hello_schema.execute("{ hello hellos helloField }")
    assert not executed.errors
    assert executed.data == {
        "hello": "World",
        "hellos": ["Worlds"],
        "helloField": "Field World",
    }


def test_query_default_value():
    class MyType(ObjectType):
        field = String()

    class Query(ObjectType):
        hello = Field(MyType, default_value=MyType(field="something else!"))

    hello_schema = Schema(Query)

    executed = hello_schema.execute("{ hello { field } }")
    assert not executed.errors
    assert executed.data == {"hello": {"field": "something else!"}}


def test_query_wrong_default_value():
    class MyType(ObjectType):
        field = String()

        @classmethod
        def is_type_of(cls, root, info):
            return isinstance(root, MyType)

    class Query(ObjectType):
        hello = Field(MyType, default_value="hello")

    hello_schema = Schema(Query)

    executed = hello_schema.execute("{ hello { field } }")
    assert len(executed.errors) == 1
    assert (
        executed.errors[0].message
        == GraphQLError("Expected value of type 'MyType' but got: 'hello'.").message
    )
    assert executed.data == {"hello": None}


def test_query_default_value_ignored_by_resolver():
    class MyType(ObjectType):
        field = String()

    class Query(ObjectType):
        hello = Field(
            MyType,
            default_value="hello",
            resolver=lambda *_: MyType(field="no default."),
        )

    hello_schema = Schema(Query)

    executed = hello_schema.execute("{ hello { field } }")
    assert not executed.errors
    assert executed.data == {"hello": {"field": "no default."}}


def test_query_resolve_function():
    class Query(ObjectType):
        hello = String()

        def resolve_hello(self, info):
            return "World"

    hello_schema = Schema(Query)

    executed = hello_schema.execute("{ hello }")
    assert not executed.errors
    assert executed.data == {"hello": "World"}


def test_query_arguments():
    class Query(ObjectType):
        test = String(a_str=String(), a_int=Int())

        def resolve_test(self, info, **args):
            return json.dumps([self, args], separators=(",", ":"))

    test_schema = Schema(Query)

    result = test_schema.execute("{ test }", None)
    assert not result.errors
    assert result.data == {"test": "[null,{}]"}

    result = test_schema.execute('{ test(aStr: "String!") }', "Source!")
    assert not result.errors
    assert result.data == {"test": '["Source!",{"a_str":"String!"}]'}

    result = test_schema.execute('{ test(aInt: -123, aStr: "String!") }', "Source!")
    assert not result.errors
    assert result.data in [
        {"test": '["Source!",{"a_str":"String!","a_int":-123}]'},
        {"test": '["Source!",{"a_int":-123,"a_str":"String!"}]'},
    ]


def test_query_input_field():
    class Input(InputObjectType):
        a_field = String()
        recursive_field = InputField(lambda: Input)

    class Query(ObjectType):
        test = String(a_input=Input())

        def resolve_test(self, info, **args):
            return json.dumps([self, args], separators=(",", ":"))

    test_schema = Schema(Query)

    result = test_schema.execute("{ test }", None)
    assert not result.errors
    assert result.data == {"test": "[null,{}]"}

    result = test_schema.execute('{ test(aInput: {aField: "String!"} ) }', "Source!")
    assert not result.errors
    assert result.data == {"test": '["Source!",{"a_input":{"a_field":"String!"}}]'}

    result = test_schema.execute(
        '{ test(aInput: {recursiveField: {aField: "String!"}}) }', "Source!"
    )
    assert not result.errors
    assert result.data == {
        "test": '["Source!",{"a_input":{"recursive_field":{"a_field":"String!"}}}]'
    }


def test_query_middlewares():
    class Query(ObjectType):
        hello = String()
        other = String()

        def resolve_hello(self, info):
            return "World"

        def resolve_other(self, info):
            return "other"

    def reversed_middleware(next, *args, **kwargs):
        return next(*args, **kwargs)[::-1]

    hello_schema = Schema(Query)

    executed = hello_schema.execute(
        "{ hello, other }", middleware=[reversed_middleware]
    )
    assert not executed.errors
    assert executed.data == {"hello": "dlroW", "other": "rehto"}


def test_objecttype_on_instances():
    class Ship:
        def __init__(self, name):
            self.name = name

    class ShipType(ObjectType):
        name = String(description="Ship name", required=True)

        def resolve_name(self, info):
            # Here self will be the Ship instance returned in resolve_ship
            return self.name

    class Query(ObjectType):
        ship = Field(ShipType)

        def resolve_ship(self, info):
            return Ship(name="xwing")

    schema = Schema(query=Query)
    executed = schema.execute("{ ship { name } }")
    assert not executed.errors
    assert executed.data == {"ship": {"name": "xwing"}}


def test_big_list_query_benchmark(benchmark):
    big_list = range(10000)

    class Query(ObjectType):
        all_ints = List(Int)

        def resolve_all_ints(self, info):
            return big_list

    hello_schema = Schema(Query)

    big_list_query = partial(hello_schema.execute, "{ allInts }")
    result = benchmark(big_list_query)
    assert not result.errors
    assert result.data == {"allInts": list(big_list)}


def test_big_list_query_compiled_query_benchmark(benchmark):
    big_list = range(100000)

    class Query(ObjectType):
        all_ints = List(Int)

        def resolve_all_ints(self, info):
            return big_list

    hello_schema = Schema(Query)
    graphql_schema = hello_schema.graphql_schema
    source = Source("{ allInts }")
    query_ast = parse(source)

    big_list_query = partial(execute, graphql_schema, query_ast)
    result = benchmark(big_list_query)
    assert not result.errors
    assert result.data == {"allInts": list(big_list)}


def test_big_list_of_containers_query_benchmark(benchmark):
    class Container(ObjectType):
        x = Int()

    big_container_list = [Container(x=x) for x in range(1000)]

    class Query(ObjectType):
        all_containers = List(Container)

        def resolve_all_containers(self, info):
            return big_container_list

    hello_schema = Schema(Query)

    big_list_query = partial(hello_schema.execute, "{ allContainers { x } }")
    result = benchmark(big_list_query)
    assert not result.errors
    assert result.data == {"allContainers": [{"x": c.x} for c in big_container_list]}


def test_big_list_of_containers_multiple_fields_query_benchmark(benchmark):
    class Container(ObjectType):
        x = Int()
        y = Int()
        z = Int()
        o = Int()

    big_container_list = [Container(x=x, y=x, z=x, o=x) for x in range(1000)]

    class Query(ObjectType):
        all_containers = List(Container)

        def resolve_all_containers(self, info):
            return big_container_list

    hello_schema = Schema(Query)

    big_list_query = partial(hello_schema.execute, "{ allContainers { x, y, z, o } }")
    result = benchmark(big_list_query)
    assert not result.errors
    assert result.data == {
        "allContainers": [
            {"x": c.x, "y": c.y, "z": c.z, "o": c.o} for c in big_container_list
        ]
    }


def test_big_list_of_containers_multiple_fields_custom_resolvers_query_benchmark(
    benchmark,
):
    class Container(ObjectType):
        x = Int()
        y = Int()
        z = Int()
        o = Int()

        def resolve_x(self, info):
            return self.x

        def resolve_y(self, info):
            return self.y

        def resolve_z(self, info):
            return self.z

        def resolve_o(self, info):
            return self.o

    big_container_list = [Container(x=x, y=x, z=x, o=x) for x in range(1000)]

    class Query(ObjectType):
        all_containers = List(Container)

        def resolve_all_containers(self, info):
            return big_container_list

    hello_schema = Schema(Query)

    big_list_query = partial(hello_schema.execute, "{ allContainers { x, y, z, o } }")
    result = benchmark(big_list_query)
    assert not result.errors
    assert result.data == {
        "allContainers": [
            {"x": c.x, "y": c.y, "z": c.z, "o": c.o} for c in big_container_list
        ]
    }


def test_query_annotated_resolvers():
    context = Context(key="context")

    class Query(ObjectType):
        annotated = String(id=String())
        context = String()
        info = String()

        def resolve_annotated(self, info, id):
            return f"{self}-{id}"

        def resolve_context(self, info):
            assert isinstance(info.context, Context)
            return f"{self}-{info.context.key}"

        def resolve_info(self, info):
            assert isinstance(info, ResolveInfo)
            return f"{self}-{info.field_name}"

    test_schema = Schema(Query)

    result = test_schema.execute('{ annotated(id:"self") }', "base")
    assert not result.errors
    assert result.data == {"annotated": "base-self"}

    result = test_schema.execute("{ context }", "base", context=context)
    assert not result.errors
    assert result.data == {"context": "base-context"}

    result = test_schema.execute("{ info }", "base")
    assert not result.errors
    assert result.data == {"info": "base-info"}


def test_default_as_kwarg_to_NonNull():
    # Related to https://github.com/graphql-python/graphene/issues/702
    class User(ObjectType):
        name = String()
        is_admin = NonNull(Boolean, default_value=False)

    class Query(ObjectType):
        user = Field(User)

        def resolve_user(self, *args, **kwargs):
            return User(name="foo")

    schema = Schema(query=Query)
    expected = {"user": {"name": "foo", "isAdmin": False}}
    result = schema.execute("{ user { name isAdmin } }")

    assert not result.errors
    assert result.data == expected
